# Copyright (C) 2025 Embedl AB

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Self, TypeAlias
from urllib.parse import urljoin

import requests
from pydantic import BaseModel
from pydantic.alias_generators import to_camel


@dataclass
class ApiConfig:
    """Configuration for interacting with the Embedl Hub REST API."""

    base_url: str
    api_key: str

    @property
    def headers(self) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }


class Model(BaseModel):
    """Base model with camel case aliasing."""

    class Config:
        alias_generator = to_camel
        validate_by_name = True


class Project(Model):
    id: str
    name: str


class Experiment(Model):
    id: str
    name: str


class RunType(Enum):
    QUANTIZE = "QUANTIZE"
    COMPILE = "COMPILE"
    BENCHMARK = "BENCHMARK"


class RunStatus(Enum):
    RUNNING = "RUNNING"
    SCHEDULED = "SCHEDULED"
    FINISHED = "FINISHED"
    FAILED = "FAILED"
    KILLED = "KILLED"


class Run(Model):
    id: str
    name: str
    type: RunType
    status: RunStatus
    created_at: datetime
    started_at: datetime
    ended_at: datetime | None


class Parameter(Model):
    name: str
    value: str
    measured_at: datetime


class Metric(Model):
    name: str
    value: float
    step: int
    measured_at: datetime


CompletedRunStatus: TypeAlias = Literal[
    RunStatus.FINISHED, RunStatus.KILLED, RunStatus.FAILED
]


JSONObj: TypeAlias = Dict[str, Any]
JSONItems: TypeAlias = List[JSONObj]
JSONData: TypeAlias = JSONObj | JSONItems


def create_project(config: ApiConfig, name: str) -> Project:
    """Create a new project."""

    data = _request(config, "POST", "/api/projects", json={"name": name})

    data = _expect_dict(data)

    return Project(**data)


def get_project_by_name(config: ApiConfig, name: str) -> Project | None:
    """Get project by name, or None if not found."""

    try:
        data = _request(
            config,
            "GET",
            "/api/projects",
            params={"name": name},
        )
    except ApiError as err:
        if err.status_code == 404:
            return None
        raise

    data = _expect_dict(data)

    return Project(**data)


def create_experiment(
    config: ApiConfig, name: str, project_id: str
) -> Experiment:
    """Create a new experiment."""

    data = _request(
        config,
        "POST",
        f"/api/projects/{project_id}/experiments",
        json={"name": name},
    )

    data = _expect_dict(data)

    return Experiment(**data)


def get_experiment_by_name(
    config: ApiConfig, name: str, project_id: str
) -> Experiment | None:
    """Get experiment by name, or None if not found."""

    try:
        data = _request(
            config,
            "GET",
            f"/api/projects/{project_id}/experiments",
            params={"name": name},
        )
    except ApiError as err:
        if err.status_code == 404:
            return None
        raise

    data = _expect_dict(data)

    return Experiment(**data)


def create_run(
    config: ApiConfig,
    project_id: str,
    experiment_id: str,
    type: RunType,
    started_at: datetime,
    name: str | None = None,
) -> Run:
    """Create a new run."""

    payload = {"type": type.value}
    if name:
        payload["name"] = name
    payload["startedAt"] = started_at.isoformat()

    data = _request(
        config,
        "POST",
        f"/api/projects/{project_id}/experiments/{experiment_id}/runs",
        json=payload,
    )

    data = _expect_dict(data)

    return Run(**data)


def update_run(
    config: ApiConfig,
    project_id: str,
    experiment_id: str,
    run_id: str,
    status: CompletedRunStatus,
    ended_at: datetime,
) -> None:
    """Update run status and end time."""

    _request(
        config,
        "PATCH",
        f"/api/projects/{project_id}/experiments/{experiment_id}/runs/{run_id}",
        json={"status": status.value, "endedAt": ended_at.isoformat()},
    )


def log_param(
    config: ApiConfig,
    name: str,
    value: str,
    project_id: str,
    experiment_id: str,
    run_id: str,
) -> Parameter:
    """Log a parameter for a run."""

    payload = {"name": name, "value": value}
    data = _request(
        config,
        "POST",
        f"/api/projects/{project_id}/experiments/{experiment_id}/runs/{run_id}/params",
        json=payload,
    )

    data = _expect_dict(data)

    return Parameter(**data)


def log_metric(
    config: ApiConfig,
    name: str,
    value: float,
    project_id: str,
    experiment_id: str,
    run_id: str,
    step: int | None = None,
) -> Metric:
    """Log a metric for a run."""

    payload = {"name": name, "value": value}
    if step is not None:
        payload["step"] = step

    data = _request(
        config,
        "POST",
        f"/api/projects/{project_id}/experiments/{experiment_id}/runs/{run_id}/metrics",
        json=payload,
    )

    data = _expect_dict(data)

    return Metric(**data)


def _expect_dict(data: JSONData | None) -> JSONObj:
    """Ensure data is a dict, else raise error."""

    if not isinstance(data, dict):
        raise RuntimeError("Unexpected response shape: expected object")
    return data


def _request(
    config: ApiConfig,
    method: str,
    url: str,
    json: JSONObj | None = None,
    params: Dict[str, str | int] | None = None,
) -> JSONData | None:
    """Send HTTP request and handle API response."""

    full_url = urljoin(config.base_url, url)

    try:
        resp = requests.request(
            method=method,
            url=full_url,
            headers=config.headers,
            json=json,
            params=params,
            timeout=10,
        )
    except requests.exceptions.RequestException as exc:
        raise NetworkRequestError(
            f"Request to {full_url} failed: {exc}"
        ) from exc

    try:
        payload: JSONObj = resp.json() if resp.content else {}
    except ValueError:
        payload = {}

    if resp.ok:
        if resp.status_code == 204:
            return None

        if "data" in payload:
            return payload["data"]

        raise ApiError(resp.status_code, "Missing `data` field", [], resp)

    errors = [
        ApiErrorDetail.from_dict(err) for err in payload.get("errors", [])
    ]
    message = payload.get("message") or resp.reason
    raise ApiError(resp.status_code, message, errors, resp)


@dataclass
class ApiErrorDetail:
    """An individual error contained in an API error response."""

    title: str | None = None
    status: int | None = None
    code: str | None = None
    detail: str | None = None
    source: dict[str, str] | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(
            title=data.get("title"),
            status=data.get("status"),
            code=data.get("code"),
            detail=data.get("detail"),
            source=data.get("source"),
        )

    def __str__(self) -> str:
        parts: list[str] = []
        if self.title:
            parts.append(self.title)
        if self.code:
            parts.append(f"(code: {self.code})")
        if self.status:
            parts.append(f"[{self.status}]")
        if self.detail:
            parts.append(self.detail)
        if self.source:
            parts.append(f"-> {self.source}")
        return " ".join(parts) or "<empty>"


class ApiError(Exception):
    """Raised for API errors with JSON body."""

    def __init__(
        self,
        status_code: int,
        message: str,
        errors: list[ApiErrorDetail] | None = None,
        response: requests.Response | None = None,
    ) -> None:
        super().__init__(f"{status_code} {message}")
        self.status_code = status_code
        self.message = message
        self.errors = errors or []
        self.response = response

    def __str__(self) -> str:
        if not self.errors:
            return super().__str__()
        joined = "\n".join(map(str, self.errors))
        return f"{super().__str__()}\n{joined}"


class NetworkRequestError(Exception):
    """Raised when HTTP request fails."""



# Copyright (C) 2025 Embedl AB

"""Utility functions for tracking experiments and runs."""

from contextlib import contextmanager

from embedl_hub.core.context import read_embedl_hub_context
from embedl_hub.core.hub_logging import console
from embedl_hub.tracking import (
    RunType,
    set_experiment,
    set_project,
    start_run,
)
from embedl_hub.tracking.utils import (
    timestamp_id,
    to_run_url,
)


def set_new_project(ctx: dict, project: str | None = None) -> None:
    """Set a new project in the context dict, using the given name or a generated one."""
    project_name = project or timestamp_id("project")
    project_ctx = set_project(project_name)
    ctx["project_id"] = project_ctx.id
    ctx["project_name"] = project_ctx.name


def set_new_experiment(ctx: dict, experiment: str | None = None) -> None:
    """Set a new experiment in the context dict, using the given name or a generated one."""
    experiment_name = experiment or timestamp_id("experiment")
    experiment_ctx = set_experiment(experiment_name)
    ctx["experiment_id"] = experiment_ctx.id
    ctx["experiment_name"] = experiment_ctx.name


def _embed_run_hyperlink(run_url: str, text: str) -> str:
    """Return a hyperlink string if the terminal supports it."""
    return (
        f"[blue][link={run_url}]{text}[/link][/]"
        if console.is_terminal
        else run_url
    )


@contextmanager
def experiment_context(
    project_name: str, experiment_name: str, run_type: RunType
):
    """
    Context manager for managing the current experiment context.
    """
    try:
        project = set_project(project_name)
        experiment = set_experiment(experiment_name)

        console.log(f"Running command with project name: {project_name}")
        console.log(f"Running command with experiment name: {experiment_name}")
        with start_run(type=run_type) as run:
            run_url = to_run_url(
                project_id=project.id,
                experiment_id=experiment.id,
                run_id=run.id,
            )
            console.log(
                f"Track your progress {_embed_run_hyperlink(run_url, 'here')}."
            )
            yield
            console.log(
                f"View results {_embed_run_hyperlink(run_url, 'here')}"
            )
    finally:
        pass



# Copyright (C) 2025 Embedl AB

"""Utility functions for the Embedl Hub CLI."""

from typing import Any, Dict, Optional, Tuple

import typer

from embedl_hub.core.hub_logging import console
from embedl_hub.tracking import global_client


def remove_none_values(input_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Remove keys with None values from a dictionary."""
    return {key: val for key, val in input_dict.items() if val is not None}


def assert_api_config():
    """Assert that the API configuration can be accessed without error."""
    try:
        _ = global_client.api_config
    except RuntimeError as e:
        console.print(f"[red]✗[/] API configuration error: {e}")
        raise typer.Exit(1)


def prepare_input_size(size: Optional[str]) -> Optional[Tuple[int, ...]]:
    """Prepare the input size from a string in comma separated format."""
    if not size:
        return None
    try:
        sizes = tuple(map(int, size.split(",")))
    except ValueError as error:
        raise ValueError(
            "Invalid size format. Use dim0, dim1,..., e.g. 1,3,224,224"
        ) from error
    console.print(f"[yellow]Using input size: {size}[/]")
    return sizes



# Copyright (C) 2025 Embedl AB
"""
Core compile logic for the embedl-hub CLI.

Handles submitting models to Qualcomm AI Hub for compilation and saving the resulting artifacts.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Tuple, Union

import qai_hub as hub

from embedl_hub.core.hardware.qualcomm_ai_hub import create_device
from embedl_hub.core.onnx_utils import maybe_package_onnx_folder_to_file
from embedl_hub.core.qai_hub_utils import (
    get_global_qai_hub_client,
    get_job_result,
    parse_runtime_info,
)
from embedl_hub.core.tracking_utils import experiment_context
from embedl_hub.tracking import RunType, log_param


class CompileError(RuntimeError):
    """Raised when Qualcomm AI Hub compile job fails or times out."""


@dataclass
class CompileResult:
    """Result of a successful compile job."""

    model_path: Path  # local .tflite, .bin, or .onnx after compile
    job_id: str
    device: str


def compile_model(
    project_name: str,
    experiment_name: str,
    model_path: Path | str,
    device: str,
    runtime: str,
    quantize_io: bool = False,
    output_file: Optional[Union[Path, str]] = None,
    input_size: Tuple[int, ...] | None = None,
) -> CompileResult:
    """
    Submit an ONNX model to Qualcomm AI Hub and retrieve the compiled artifact.

    Args:
        model_path: Path to TorchScript/ONNX (INT8 or FP32).
        device: Device nickname, e.g. 'Samsung Galaxy S24'.
        runtime: 'tflite' | 'qnn' | 'onnx'
        quantize_io: Add --quantize_io to options.
        output_file: Compiled model filename. File ending added automatically if not specified.
        input_size: Input size as (dim0, dim1, ...) tuple, e.g. (1, 3, 224, 224).

    Returns:
        CompileResult with local path to compiled model.
    """

    model_path = Path(model_path)
    if not model_path.exists():
        raise ValueError(f"Model not found: {model_path}")

    with experiment_context(project_name, experiment_name, RunType.COMPILE):
        with TemporaryDirectory() as tmpdir:
            model_path = maybe_package_onnx_folder_to_file(model_path, tmpdir)
            return _compile_model(
                model_file=model_path,
                device=device,
                runtime=runtime,
                quantize_io=quantize_io,
                output_file=output_file,
                input_size=input_size,
            )


def _compile_model(
    model_file: Path,
    device: str,
    runtime: str = "tflite",
    quantize_io: bool = False,
    output_file: Optional[Union[Path, str]] = None,
    input_size: Tuple[int, ...] | None = None,
) -> CompileResult:
    """
    Submit a model to Qualcomm AI Hub and retrieve the compiled artifact.

    Args:
        model_file: Path to the model.
        device: Device nickname, e.g. 'Samsung Galaxy S24'.
        runtime: 'tflite' | 'qnn' | 'onnx'
        quantize_io: Add --quantize_io to options.
        output_file: Compiled model filename. File ending added automatically if not specified.
        input_size: Input size as tuple, e.g. (1, 3, 224, 224).

    Returns:
        CompileResult with local path to compiled model.

    """
    hub_device = create_device(device)

    log_param("$device", device)

    opts = f"--target_runtime {runtime}"
    if quantize_io:
        opts += " --quantize_io"

    input_specs = {"image": input_size} if input_size else None

    try:
        job = hub.submit_compile_job(
            model=model_file.as_posix(),
            device=hub_device,
            options=opts,
            input_specs=input_specs,
        )
    except Exception as error:
        raise CompileError("Failed to submit compile job.") from error

    log_param("$qai_hub_job_id", job.job_id)

    try:
        compiled_model = job.get_target_model()
    except Exception as error:
        raise CompileError(
            "Failed to download compiled model from Qualcomm AI Hub."
        ) from error
    local_path = _save_compiled_model(compiled_model, output_file)

    if "qnn" in runtime:
        # Compile jobs with --target-runtime qnn_dlc, qnn_lib_aarch64_android or
        # qnn_context_binary don't expose runtime info in the job result
        logged_runtime = "QNN"
    else:
        try:
            job_result = get_job_result(
                job.job_id, get_global_qai_hub_client().config
            )
            logged_runtime = parse_runtime_info(job_result)
        except RuntimeError as error:
            raise CompileError("Failed to parse runtime info.") from error
    log_param("$runtime", logged_runtime)

    return CompileResult(
        model_path=local_path,
        job_id=job.job_id,
        device=device,
    )


def _save_compiled_model(
    compiled: "hub.Model", output_file: Optional[Union[Path, str]] = None
) -> Path:
    """
    Save the compiled artifact to a user-specified location or to the root
    directory with the name provided by Qualcomm AI Hub.

    Returns:
        Path to the saved model file.
    """
    model_filename = Path(compiled.name)
    if output_file is None:
        dst = Path.cwd() / model_filename
    else:
        output_path = Path(output_file)
        if output_path.is_dir() or str(output_file).endswith(("/", "\\")):
            dst = output_path / model_filename
        else:
            dst = output_path

    dst.parent.mkdir(parents=True, exist_ok=True)
    dst = dst.with_suffix("")
    dst = compiled.download(str(dst))
    return dst



# Copyright (C) 2025 Embedl AB

"""
Functionality for handling the Qualcomm AI Hub communication and device management.
"""

import qai_hub as hub
from rich.table import Table

from embedl_hub.core.hub_logging import console


def print_device_table():
    """Print a table of available devices on the Qualcomm AI Hub."""
    table = Table(title="Embedl Hub devices")
    table.add_column("Name", style="cyan")
    table.add_column("Cloud")
    table.add_column("OS")
    table.add_column("Chipset / Attrs", overflow="fold")
    devices = [
        (
            d.name,
            "qai_hub",
            d.os or "—",
            ", ".join(a for a in d.attributes if "chipset" in a) or "—"
        )
        for d in hub.get_devices()
    ]
    for row in devices:
        table.add_row(*row)
    console.print(table)


def create_device(device_name: str) -> hub.Device:
    """
    Create a device object from the given device name.

    Args:
        device (str): The name of the device.

    Returns:
        hub.Device: The created device object.
    """
    all_device_names = [d.name for d in hub.get_devices()]
    if device_name not in all_device_names:
        raise ValueError(
            f"Device name '{device_name}' not recognized. "
            "Use `embedl-hub list-devices` to see available devices."
        )
    return hub.Device(device_name)



#  Copyright (c) 2018-2024, Qualcomm Innovation Center, Inc. All rights reserved.
#  Copyright (C) 2025 Embedl AB.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
#
#  SPDX-License-Identifier: BSD-3-Clause

import numpy as np
from aimet_common.utils import AimetLogger


def compute_psnr(
    expected: np.ndarray,
    actual: np.ndarray,
    max_psnr: float = 100.0,
) -> float:
    """
    Computes the Peak Signal-to-Noise Ratio (PSNR) for two values
    PSNR = 20 * log10(data_range / noise) where noise is Root Mean Square error between expected and actual

    Where the data_range is the dynamic range of expected value, computed as maximum of:

    1. expected.max() - expected.min() (standard dynamic range)
    2. np.abs(expected).max() (absolute max value)

    NOTE: Ensure PSNR between (-100, 100) dB to ensure numerical stability in downstream tasks like sorting during per-layer analysis.
          Uses separate thresholds for data_range and noise to ensure smooth behaviour near zero.

    :param expected: The reference values or ground truth (FP32 outputs)
    :param actual: The noisy values to compare against (quantized outputs)
    :param max_psnr: The maximum PSNR value to clip to (default is 100 dB)
    :return: The computed PSNR value in dB
    """
    logger = AimetLogger.get_area_logger(AimetLogger.LogAreas.Utils)

    expected = np.asarray(expected)
    actual = np.asarray(actual)

    if expected.shape != actual.shape:
        raise ValueError("Input arrays must have the same shape.")

    if expected.size == 0 or actual.size == 0:
        raise ValueError("Input arrays must not be empty.")

    # Allow only finite expected values (Exclude NaN, +inf and -inf)
    if not np.all(np.isfinite(expected)):
        raise ValueError("Input arrays must have finite values.")

    # If the actual values contain NaN or +inf or -inf, treat as worst-case quality by returning -max_psnr
    if not np.all(np.isfinite(actual)):
        logger.warning("The actual values contain NaNs or infinite values.")
        return -max_psnr

    # Compute maximum custom data range
    epsilon = 1e-10  # The minimum value for data range to avoid instability
    data_range = max(expected.max() - expected.min(), np.abs(expected).max())
    data_range = max(data_range, epsilon)  # Avoid log10(very small / noise)

    # Compute noise power and RMS error
    noise_epsilon = epsilon * 10 ** (-max_psnr / 20)
    noise_pw = np.mean(np.square(expected - actual))
    noise = max(
        np.sqrt(noise_pw), noise_epsilon
    )  # Avoid log10(data_range / 0)

    # Compute PSNR and clip to ensure numerical stability
    psnr = 20 * np.log10(data_range / noise)
    return float(np.clip(psnr, -max_psnr, max_psnr))



# Copyright (C) 2025 Embedl AB

"""
Public Embedl Hub library API.
```pycon
>>> import embedl_hub
>>> embedl_hub.__version__
'2025.9.0'
```
"""

import importlib.metadata


try:
    __version__ = importlib.metadata.version("embedl_hub")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"



# Copyright (C) 2025 Embedl AB

"""Tests for the functionality of the ExperimentConfig class."""

import pytest
import yaml

from embedl_hub.core.config import ExperimentConfig


class MainConfig(ExperimentConfig):
    """Main configuration class for testing. Contains an inner config."""

    a: str = "hello"
    x: int = 3
    y: int = 2


def test_from_dict_and_to_dict_roundtrip():
    """Test the round-trip conversion of a dictionary to a config object and back."""
    data = {"a": "world", "x": 10}
    cfg = MainConfig(**data)
    assert cfg.a == "world"
    assert cfg.x == 10
    assert cfg.y == 2

    # round-trip via to_dict()
    output = cfg.model_dump()
    assert output == {"a": "world", "x": 10, "y": 2}


def test_merge_dict_overrides():
    """Test merging a dictionary into the config object with overrides."""
    cfg1 = MainConfig()
    # override a and y only
    cfg2 = cfg1.merge_dict({"a": "hello world", "y": 5})
    assert cfg2.a == "hello world"  # overridden
    assert cfg2.y == 5
    assert cfg1.y == 2  # original unchanged

    # override via kwargs
    cfg3 = cfg1.merge_dict({}, a="new", x=7)
    assert cfg3.a == "new"
    assert cfg3.x == 7


def test_merge_yaml(tmp_path):
    """Test merging a YAML file into the config object."""
    # create a temporary YAML file
    file = tmp_path / "cfg.yaml"
    yaml.dump({"a": "tmp", "y": 9}, file.open("w"))
    cfg1 = MainConfig()
    cfg4 = cfg1.merge_yaml(file)
    assert cfg4.a == "tmp"
    assert cfg4.y == 9


def test_merge_yaml_with_kwargs(tmp_path):
    """Test merging a YAML file into the config object with kwargs."""
    # create a temporary YAML file
    file = tmp_path / "cfg.yaml"
    yaml.dump({"a": "file", "x": 3, "y": 4}, file.open("w"))
    cfg1 = MainConfig()
    # merge YAML and override with kwargs
    cfg5 = cfg1.merge_yaml(file, a="override", y=99)
    assert cfg5.a == "override"  # kwarg overrides YAML
    assert cfg5.x == 3  # from file
    assert cfg5.y == 99  # kwarg overrides y


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2025 Embedl AB

"""
embedl-hub quantize - send an onnx model to Qualcomm AI Hub and retrieve a
quantized onnx model.
"""

from pathlib import Path

import typer

# All other embedl_hub imports should be done inside the function.
from embedl_hub.cli.helper import (
    CONFIG_HELPER,
    OUTPUT_FILE_HELPER,
)

quantize_cli = typer.Typer()


@quantize_cli.command("quantize")
def quantize_command(
    model: Path = typer.Option(
        ...,
        "-m",
        "--model",
        help="Path to the ONNX model file, or to a directory containing the ONNX model "
        "and any associated data files, to be quantized. (required)",
        show_default=False,
    ),
    data_path: Path = typer.Option(
        None,
        "--data",
        "-d",
        help=(
            "Path to the dataset used for calibration. "
            "If not provided, random data will be used for calibration. "
            "If the model is a single-input model, the directory should contain "
            "numpy files (.npy) with the input data. "
            "If the model has multiple inputs, the directory should contain "
            "subdirectories named after the input names, each containing "
            "numpy files (.npy) with the corresponding input data."
        ),
        show_default=False,
    ),
    output_file: Path = typer.Option(
        None,
        "-o",
        "--output-file",
        help=OUTPUT_FILE_HELPER,
        show_default=False,
    ),
    num_samples: int = typer.Option(
        None,
        "--num-samples",
        "-n",
        help="Number of data samples to use during quantization calibration.",
        show_default=False,
    ),
    config_path: Path = typer.Option(
        None,
        "--config",
        "-c",
        help=CONFIG_HELPER,
    ),
):
    """
    Quantize an ONNX model using Qualcomm AI Hub.
    Qualcomm AI Hub may return a zip file containing multiple files.

    Required arguments (if not provided through --config):
        --model

    Examples
    --------
    Quantize the ONNX model `compiled_model.onnx` calibrating on data
    from `/path/to/dataset/` using 1000 samples from the dataset:

        $ embedl-hub quantize -m compiled_model.onnx -d /path/to/dataset/ -n 1000

    Quantize the ONNX model using the configuration specified in `my_config.yaml`:

        $ embedl-hub quantize -c ./my_config.yaml

    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import assert_api_config, remove_none_values
    from embedl_hub.core.config import load_default_config_with_size
    from embedl_hub.core.context import require_embedl_hub_context
    from embedl_hub.core.hub_logging import console
    from embedl_hub.core.quantization.quantization_config import (
        QuantizationConfig,
    )
    from embedl_hub.core.quantization.quantize import (
        quantize_model,
    )
    # pylint: enable=import-outside-toplevel

    assert_api_config()
    ctx = require_embedl_hub_context()

    if not output_file:
        output_file = Path(model).with_suffix(".quantized")
        console.print(
            f"[yellow]No output file specified, using default: {output_file}[/]"
        )

    cfg = load_default_config_with_size(QuantizationConfig, None, "quantize")
    cli_flags = remove_none_values(
        {
            "model": model,
            "output_file": output_file,
            "num_samples": num_samples,
            "data_path": data_path,
        }
    )
    cfg = cfg.merge_yaml(other=config_path, **cli_flags)
    try:
        cfg.validate_config()
    except ValueError as e:
        console.print(f"[red]Error: {e}[/]")
        raise typer.Exit(1)

    quantize_model(
        config=cfg,
        project_name=ctx["project_name"],
        experiment_name=ctx["experiment_name"],
    )



# Copyright (C) 2025 Embedl AB
"""
Project and experiment context management for embedl-hub CLI.

This module provides CLI commands to initialize and display the current project
and experiment context. The selected context determines under which project and
experiment all data, results, and metadata are stored in the user's account on
https://hub.embedl.com.

Users can create new projects and experiments, switch between them, and view the
active context. Context information is stored locally in a YAML file.
"""

from __future__ import annotations

from typing import Optional

import typer
from rich.table import Table

init_cli = typer.Typer(help="Initialise / show project & experiment context")


@init_cli.command("init")
def init_command(
    project: Optional[str] = typer.Option(
        None, "-p", "--project", help="Project name or id", show_default=False
    ),
    experiment: Optional[str] = typer.Option(
        None,
        "-e",
        "--experiment",
        help="Experiment name or id",
        show_default=False,
    ),
):
    """
    Configure persistent CLI context.

    This command stores values used by other commands in a local context file
    in your home directory. The context file can contain:

    - Active project (created automatically if name does not exist)
    - Active experiment (created automatically if name does not exist)

    Examples
    --------
    Create a new project and experiment with random names:
        $ embedl-hub init

    Create or set named project:
        $ embedl-hub init -p "My Flower Detector App"

    Create or load named experiment inside current project:
        $ embedl-hub init -e "MobileNet Flower Detector"

    Set both project and experiment explicitly:
        $ embedl-hub init -p "My Flower Detector App" -e "MobileNet Flower Detector"
    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import assert_api_config
    from embedl_hub.core.context import (
        read_embedl_hub_context,
        write_ctx,
    )
    from embedl_hub.core.hub_logging import console
    from embedl_hub.core.tracking_utils import (
        set_new_experiment,
        set_new_project,
    )
    from embedl_hub.tracking.utils import (
        to_experiment_url,
    )
    # pylint: enable=import-outside-toplevel

    assert_api_config()

    ctx = read_embedl_hub_context()
    current_project = ctx.get("project_name")
    current_experiment = ctx.get("experiment_name")
    force_new_experiment = project is not None and project != current_project

    # Decide which project to use
    project_to_use: Optional[str]
    if project:
        console.print(f"Setting project to '{project}'")
        project_to_use = project
    elif not current_project:
        console.print("No active project, creating a new one")
        project_to_use = project
    else:
        console.print(f"Keeping current project '{current_project}'")
        project_to_use = current_project
    set_new_project(ctx, project_to_use)

    # Decide which experiment to use
    experiment_to_use: Optional[str]
    if experiment:
        console.print(f"Setting experiment to '{experiment}'")
        experiment_to_use = experiment
    elif not current_experiment or force_new_experiment:
        console.print("No active experiment, creating a new one")
        experiment_to_use = experiment
    else:
        console.print(f"Keeping current experiment '{current_experiment}'")
        experiment_to_use = current_experiment
    set_new_experiment(ctx, experiment_to_use)

    write_ctx(ctx)

    console.print(f"[green]✓ Project:[/] {ctx['project_name']}")
    console.print(f"[green]✓ Experiment:[/] {ctx['experiment_name']}")
    console.print(
        f"See your results: {to_experiment_url(ctx['project_id'], ctx['experiment_id'])}"
    )


@init_cli.command("show")
def show_command():
    """Print active project/experiment IDs and names."""

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import assert_api_config
    from embedl_hub.core.context import read_embedl_hub_context
    from embedl_hub.core.hub_logging import console
    # pylint: enable=import-outside-toplevel

    assert_api_config()
    ctx = read_embedl_hub_context()

    if not ctx:
        console.print("[red]✗[/] No project or experiment initialized.")
        console.print(
            "Run `embedl-hub init` to create a new project and experiment."
        )
        return

    table = Table(title=".embedl_hub_ctx", show_lines=True, show_header=False)
    for k in (
        "project_id",
        "project_name",
        "experiment_id",
        "experiment_name",
    ):
        table.add_row(k, ctx.get(k, "—"))
    console.print(table)



# Copyright (C) 2025 Embedl AB

"""Test the Embedl Hub SDK CLI utils."""

import pytest

from embedl_hub.cli.utils import prepare_input_size


@pytest.mark.parametrize(
    "size_str, expected_tuple",
    [
        ("1,3,224,224", (1, 3, 224, 224)),
        ("1,28,28", (1, 28, 28)),
        ("1", (1,)),
    ],
)
def test_prepare_input_size_valid(size_str, expected_tuple, capsys):
    """Test prepare_input_size with valid comma-separated integer strings."""
    result = prepare_input_size(size_str)
    assert result == expected_tuple
    captured = capsys.readouterr()
    assert f"Using input size: {size_str}" in captured.out


def test_prepare_input_size_empty():
    """Test prepare_input_size with an empty string."""
    assert prepare_input_size("") is None


def test_prepare_input_size_invalid():
    """Test prepare_input_size with an invalid string."""
    with pytest.raises(
        ValueError,
        match="Invalid size format. Use dim0, dim1,..., e.g. 1,3,224,224",
    ):
        prepare_input_size("1,three,224,224")



# Copyright (C) 2025 Embedl AB

"""Shared Rich console and logging setup helpers for embedl-hub CLI.

Import ``console`` whenever you need to print user-facing output:

    from embedl_hub.cli.logging import console
    console.print("[green]✓ done[/]")

Call ``setup_logging(verbose)`` once at Typer app entry-point, passing the
``-v/--verbose`` count so that log levels map to WARNING/INFO/DEBUG/NOTSET.

"""

import logging
from typing import Sequence

from rich.console import Console
from rich.logging import RichHandler

__all__: Sequence[str] = ("console", "setup_logging")

# This is the single Console instance reused across the entire CLI
console: Console = Console()


def setup_logging(verbose: int = 0, *, force: bool = False) -> None:
    """Configure Rich-formatted logging.

    Parameters
    ----------
    verbose : int, default 0
        Verbosity count from ``-v/--verbose`` flag.
        0 → WARNING, 1 → INFO, 2 → DEBUG, ≥3 → NOTSET.
    force : bool, default False
        If True, re-configures logging even if already configured.
        Handy for unit tests that run the Typer app multiple times.
    """

    levels = [logging.WARNING, logging.INFO, logging.DEBUG, logging.NOTSET]
    level = levels[min(verbose, len(levels) - 1)]

    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(console=console, markup=True, rich_tracebacks=True)
        ],
        force=force,
    )

    # Here we can silence 3rd party libraries if needed by adding them to the list of names below.
    if verbose == 0:
        name: str
        for name in ():
            logging.getLogger(name).setLevel(logging.WARNING)



# Copyright (C) 2025 Embedl AB

import os
from datetime import datetime, timezone
from urllib.parse import urljoin

from embedl_hub.tracking.client import (
    BASE_URL_ENV_VAR_NAME,
    DEFAULT_API_BASE_URL,
)
from embedl_hub.tracking.rest_api import Experiment, Project, Run


def to_project_url(project_id: str):
    """Convert a Project object to its URL representation."""
    base_url = os.getenv(BASE_URL_ENV_VAR_NAME, DEFAULT_API_BASE_URL)

    return urljoin(base_url, f"projects/{project_id}")


def to_experiment_url(project_id: str, experiment_id: str):
    """Convert an Experiment object to its URL representation."""
    base_url = os.getenv(BASE_URL_ENV_VAR_NAME, DEFAULT_API_BASE_URL)
    return urljoin(
        base_url, f"projects/{project_id}/experiments/{experiment_id}"
    )


def to_run_url(project_id: str, experiment_id: str, run_id: str):
    """Convert a Run object to its URL representation."""
    base_url = os.getenv(BASE_URL_ENV_VAR_NAME, DEFAULT_API_BASE_URL)
    return urljoin(
        base_url,
        f"projects/{project_id}/experiments/{experiment_id}/runs/{run_id}",
    )


def timestamp_id(prefix: str) -> str:
    """Generate a unique random name based on timestamp."""
    return (
        f"{prefix}_{datetime.now(tz=timezone.utc).strftime('%Y%m%d_%H%M%S')}"
    )



# Copyright (C) 2025 Embedl AB

"""Context manager for managing the current experiment context."""

from pathlib import Path
from typing import Dict

import typer
import yaml

from embedl_hub.core.hub_logging import console

CTX_FILE = Path.home() / ".embedl_hub_ctx"


def write_ctx(ctx: Dict[str, str]) -> None:
    """Write the current embedl-hub context to the local YAML file."""
    CTX_FILE.write_text(yaml.safe_dump(ctx, sort_keys=False), encoding="utf-8")


def read_embedl_hub_context() -> Dict[str, str]:
    """Read the current embedl-hub context from the local YAML file."""
    return (
        yaml.safe_load(CTX_FILE.read_text(encoding="utf-8"))
        if CTX_FILE.exists()
        else {}
    )


def require_embedl_hub_context() -> Dict[str, str]:
    """
    Load the current embedl-hub context and ensure `project_name` and `experiment_name` are present.
    """
    ctx = read_embedl_hub_context()
    if not ctx.get("project_name") or not ctx.get("experiment_name"):
        console.print(
            "[red]Failed to find context: No project or experiment is initialized.[/]\n",
            "[red]Run 'embedl-hub init' to set up a project and experiment.[/]\n",
            "[red]Run 'embedl-hub auth' to set up an api key.[/]",
        )
        raise typer.Exit(1)
    return ctx



# Copyright (C) 2025 Embedl AB

"""
Test for the experiment_context manager in embedl_hub.core.context
"""

import pytest

from embedl_hub.core.hub_logging import console
from embedl_hub.core.tracking_utils import experiment_context
from embedl_hub.tracking import RunType


@pytest.fixture(autouse=True)
def mock_tracking(monkeypatch):
    """Pytest fixture to mock the tracking functions used in the experiment_context."""

    class DummyCtx:
        def __init__(self, id, name):
            self.id = id
            self.name = name

    class DummyRun:
        def __init__(self):
            self.id = "1337"
            self.name = "dummy_run"

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.set_project",
        lambda name: DummyCtx(f"dummy_project_id_{name}", name),
    )
    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.set_experiment",
        lambda name: DummyCtx(f"dummy_experiment_id_{name}", name),
    )
    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.start_run",
        lambda **kwargs: DummyRun(),
    )


def test_experiment_context_logs(capsys, monkeypatch):
    """Test that the experiment_context logs the correct messages."""

    # Set a large width to prevent rich from truncating the output
    monkeypatch.setattr(console, "width", 120)

    with experiment_context(
        experiment_name="test_experiment",
        project_name="test_project",
        run_type=RunType.QUANTIZE,
    ):
        pass

    captured = capsys.readouterr()
    messages = captured.out + captured.err

    # Check that both messages appeared
    assert "Running command with project name: test_project" in messages
    assert "Running command with experiment name: test_experiment" in messages


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2025 Embedl AB

"""Module for measuring PSNR between float and quantized ONNX models."""

import json
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Tuple

import numpy as np
import onnx
import onnxruntime as ort
from onnx import ModelProto, NodeProto, TensorProto, helper, numpy_helper

from embedl_hub.thirdparty.aimet.utils import compute_psnr


def load_model(path: str) -> ModelProto:
    """Loads an ONNX model, including external data."""
    return onnx.load(path, load_external_data=True)


def save_model(model: ModelProto, out_path: Path):
    """Saves an ONNX model, creating the directory if it doesn't exist."""
    onnx.save_model(model, out_path)


def _infer_model_shapes(model: ModelProto) -> ModelProto:
    """Infers shapes and types for all tensors in the model."""
    return onnx.shape_inference.infer_shapes(
        model, strict_mode=True, data_prop=True
    )


def _get_value_info_index(
    model: ModelProto,
) -> Dict[str, Tuple[int, List[int]]]:
    """Creates a dictionary mapping tensor names to their type and shape."""
    index = {}

    def add_value_info_to_index(value_info):
        tensor_type = value_info.type.tensor_type
        if tensor_type and tensor_type.elem_type != 0:
            dims = []
            if tensor_type.HasField("shape"):
                for dim in tensor_type.shape.dim:
                    dims.append(
                        dim.dim_value if dim.HasField("dim_value") else None
                    )
            else:
                dims = None
            index[value_info.name] = (tensor_type.elem_type, dims)

    for value_info in (
        list(model.graph.input)
        + list(model.graph.output)
        + list(model.graph.value_info)
    ):
        add_value_info_to_index(value_info)
    return index


def _make_tensors_outputs(model: ModelProto, tensor_names: List[str]):
    """
    Adds the given tensors to the model's output list.
    It infers shapes and types to create the necessary ValueInfoProto.
    """
    inferred_model = ModelProto()
    inferred_model.CopyFrom(model)
    try:
        inferred_model = _infer_model_shapes(inferred_model)
    except Exception:
        # Shape inference may fail, but we can try to proceed.
        pass

    value_info_index = _get_value_info_index(inferred_model)
    existing_outputs = {o.name for o in model.graph.output}

    for tensor_name in tensor_names:
        if tensor_name in existing_outputs:
            continue

        if tensor_name in value_info_index:
            elem_type, dims = value_info_index[tensor_name]
            value_info = helper.make_tensor_value_info(
                tensor_name, elem_type, dims if dims is not None else None
            )
        else:
            # Fallback for tensors without inferred types (common for intermediate tensors)
            value_info = helper.make_tensor_value_info(
                tensor_name, TensorProto.FLOAT, None
            )
        model.graph.output.append(value_info)


def run_inference(
    model_path: Path, feed_dict: Dict[str, np.ndarray]
) -> Dict[str, np.ndarray]:
    """
    Runs inference on an ONNX model using ONNX Runtime.
    """
    session = ort.InferenceSession(
        model_path, providers=["CUDAExecutionProvider", "CPUExecutionProvider"]
    )

    # If the model has a single input, auto-map the input name.
    if len(feed_dict) == 1 and len(session.get_inputs()) == 1:
        input_tensor = next(iter(feed_dict.values()))
        input_name = session.get_inputs()[0].name
        feed_dict = {input_name: input_tensor}

    # Validate input names
    session_input_names = {i.name for i in session.get_inputs()}
    unknown_inputs = [k for k in feed_dict if k not in session_input_names]
    if unknown_inputs:
        raise ValueError(
            f"Unknown input(s) {unknown_inputs}. Available: {sorted(session_input_names)}"
        )

    output_names = [o.name for o in session.get_outputs()]
    outputs = session.run(output_names, feed_dict)
    return dict(zip(output_names, outputs))


def get_initializer_dict(model: ModelProto) -> Dict[str, np.ndarray]:
    """Returns a dictionary of model initializers."""
    return {
        init.name: numpy_helper.to_array(init)
        for init in model.graph.initializer
    }


# --------------------
# Graph Analysis
# --------------------
def get_all_node_output_names(model: ModelProto) -> List[str]:
    """Returns a list of all unique node output tensor names in the model."""
    all_outputs = []
    for node in model.graph.node:
        all_outputs.extend([o for o in node.output if o])

    # De-duplicate while preserving order
    seen = set()
    unique_outputs = []
    for tensor_name in all_outputs:
        if tensor_name not in seen:
            seen.add(tensor_name)
            unique_outputs.append(tensor_name)
    return unique_outputs


def get_post_quantization_map(model: ModelProto) -> Dict[str, dict]:
    """
    Creates a map of tensors that are inputs to QuantizeLinear nodes.

    For each tensor `T` that is consumed by a `QuantizeLinear` node, this function
    records the output name of the `QuantizeLinear` node, and the scale and zero-point
    values used for quantization.

    Returns:
        A dictionary mapping a tensor name to its quantization info:
        {
            "tensor_name": {
                "ql_out": <output_name_of_quantize_linear>,
                "scale": <scale_value>,
                "zp": <zero_point_value>
            }
        }
    """
    initializers = get_initializer_dict(model)
    consumers: Dict[str, List[NodeProto]] = {}
    for node in model.graph.node:
        for input_name in node.input:
            consumers.setdefault(input_name, []).append(node)

    post_quantization_map = {}
    for tensor_name, consumer_nodes in consumers.items():
        quantize_linear_nodes = [
            node for node in consumer_nodes if node.op_type == "QuantizeLinear"
        ]
        if not quantize_linear_nodes:
            continue

        quant_node = quantize_linear_nodes[0]
        if len(quant_node.input) >= 3:
            scale_name, zp_name = quant_node.input[1], quant_node.input[2]
            if scale_name in initializers and zp_name in initializers:
                post_quantization_map[tensor_name] = {
                    "ql_out": quant_node.output[0],
                    "scale": initializers[scale_name].astype(np.float32),
                    "zp": initializers[zp_name],
                }
    return post_quantization_map


def add_debug_outputs(
    model: ModelProto, tensor_names: List[str]
) -> ModelProto:
    """Clones a model and adds the specified tensors to its outputs."""
    cloned_model = ModelProto()
    cloned_model.CopyFrom(model)
    _make_tensors_outputs(cloned_model, tensor_names)
    if not cloned_model.opset_import:  # Safety check
        cloned_model.opset_import.extend(model.opset_import)
    return cloned_model


def dequantize_tensor(
    quantized_tensor: np.ndarray,
    scale: np.ndarray,
    zero_point: np.ndarray,
    original_ndim: int,
) -> np.ndarray:
    """Dequantizes a tensor using its scale and zero-point."""
    if np.size(scale) == 1 and np.size(zero_point) == 1:
        return (
            quantized_tensor.astype(np.float32) - float(zero_point)
        ) * float(scale)

    # Handle per-channel quantization (common in NCHW format)
    if original_ndim >= 2 and (
        scale.size == quantized_tensor.shape[1]
        or np.size(zero_point) == quantized_tensor.shape[1]
    ):
        # Reshape scale and zero-point for broadcasting
        broadcast_shape = (1, -1) + (1,) * (original_ndim - 2)
        scale_b = np.array(scale, dtype=np.float32).reshape(broadcast_shape)
        zp_b = np.array(zero_point, dtype=np.float32).reshape(broadcast_shape)
        return (quantized_tensor.astype(np.float32) - zp_b) * scale_b

    # Fallback for other cases
    return (
        quantized_tensor.astype(np.float32)
        - np.array(zero_point, dtype=np.float32)
    ) * np.array(scale, dtype=np.float32)


def _prepare_debug_model(
    model_path: str,
    tmp_folder: Path,
    extra_output_names: Optional[List[str]] = None,
) -> Tuple[Path, List[str]]:
    """Prepares a model for debug inference by adding all tensor outputs."""
    model = load_model(model_path)
    tensor_names = get_all_node_output_names(model)
    all_outputs = list(set(tensor_names + (extra_output_names or [])))
    debug_model = add_debug_outputs(model, all_outputs)
    debug_model_path = tmp_folder / f"{Path(model_path).stem}_debug.onnx"
    save_model(debug_model, debug_model_path)
    return debug_model_path, tensor_names


def _calculate_psnr_for_tensors(
    float_tensor_names: List[str],
    float_outputs: Dict[str, np.ndarray],
    qdq_outputs: Dict[str, np.ndarray],
    post_quantization_map: Dict[str, dict],
) -> List[dict]:
    """Calculates PSNR for corresponding float and dequantized tensors."""
    results = []
    for tensor_name in float_tensor_names:
        if (
            tensor_name not in float_outputs
            or tensor_name not in post_quantization_map
        ):
            continue

        quant_info = post_quantization_map[tensor_name]
        ql_out_name = quant_info["ql_out"]
        if ql_out_name not in qdq_outputs:
            continue

        float_tensor = float_outputs[tensor_name]
        quantized_tensor = qdq_outputs[ql_out_name]
        scale, zero_point = quant_info["scale"], quant_info["zp"]
        dequantized_tensor = dequantize_tensor(
            quantized_tensor, scale, zero_point, float_tensor.ndim
        )

        if float_tensor.shape != dequantized_tensor.shape:
            continue

        results.append(
            {
                "layer": tensor_name,
                "shape": tuple(float_tensor.shape),
                "psnr": float(compute_psnr(float_tensor, dequantized_tensor)),
            }
        )
    return results


# --------------------
# Main Comparison Logic
# --------------------
def compare_models(
    float_model_path: str,
    qdq_model_path: str,
    inputs: Dict[str, np.ndarray],
) -> Tuple[List[dict], List[dict]]:
    """
    Compares the layer and model outputs of a float model and a QDQ model.

    This function calculates the PSNR between the float tensor output of each layer
    and the dequantized output of the corresponding layer in the QDQ model. It
    also computes the PSNR for the final model outputs.

    Returns:
        A tuple containing:
        - A list of dictionaries for layer-wise PSNR results.
        - A list of dictionaries for model output PSNR results.
    """
    float_model = load_model(float_model_path)
    qdq_model = load_model(qdq_model_path)
    post_quantization_map = get_post_quantization_map(qdq_model)
    extra_qdq_outputs = [
        info["ql_out"] for info in post_quantization_map.values()
    ]

    with TemporaryDirectory() as tmpdir:
        tmp_folder = Path(tmpdir)

        float_debug_path, float_tensor_names = _prepare_debug_model(
            float_model_path, tmp_folder
        )
        qdq_debug_path, _ = _prepare_debug_model(
            qdq_model_path, tmp_folder, extra_qdq_outputs
        )

        float_outputs = run_inference(float_debug_path, inputs)
        qdq_outputs = run_inference(qdq_debug_path, inputs)

    # Calculate PSNR for intermediate layers
    layer_results = _calculate_psnr_for_tensors(
        float_tensor_names, float_outputs, qdq_outputs, post_quantization_map
    )

    # Calculate PSNR for model outputs
    float_output_names = [o.name for o in float_model.graph.output]
    qdq_output_names = [o.name for o in qdq_model.graph.output]
    output_results = []

    for float_out_name, qdq_out_name in zip(
        float_output_names, qdq_output_names
    ):
        if float_out_name in float_outputs and qdq_out_name in qdq_outputs:
            float_tensor = float_outputs[float_out_name]
            qdq_tensor = qdq_outputs[qdq_out_name]

            if float_tensor.shape != qdq_tensor.shape:
                output_results.append(
                    {
                        "layer": f"{float_out_name} vs {qdq_out_name}",
                        "shape": "NA",
                        "psnr": None,
                        "note": f"Shape mismatch: {float_tensor.shape} vs {qdq_tensor.shape}",
                    }
                )
                continue

            output_results.append(
                {
                    "layer": float_out_name,
                    "shape": tuple(float_tensor.shape),
                    "psnr": float(compute_psnr(float_tensor, qdq_tensor)),
                }
            )

    return layer_results, output_results


def print_comparison_results(title: str, results: List[dict]):
    """Prints the comparison results in a formatted table."""
    print(f"\n=== {title} ===")
    for result in results:
        shape = result["shape"]
        shape_str = str(shape) if isinstance(shape, tuple) else "NA"
        psnr_str = "NA" if result["psnr"] is None else f"{result['psnr']:.3f}"
        note = result.get("note", "")
        print(
            f"{result['layer']:<40} psnr={psnr_str:>8}  shape={shape_str}  {note}"
        )


def measure_psnr_between_models(
    float_model_path: str,
    qdq_model_path: str,
    calibration_data: Dict[str, List[np.ndarray]],
):
    """
    Measures and prints the PSNR between a float model and a QDQ model.
    It computes PSNR for both intermediate layer outputs and final model outputs.

    Args:
        float_model_path: Path to the float ONNX model.
        qdq_model_path: Path to the QDQ ONNX model.
        inputs: Dictionary of input tensors for inference.
        title: Title for the comparison results.
    """
    inputs = {
        k: np.concatenate(v, axis=0) if len(v) > 1 else v[0]
        for k, v in calibration_data.items()
    }
    layer_results, output_results = compare_models(
        float_model_path, qdq_model_path, inputs
    )
    print_comparison_results("Layerwise PSNR", layer_results)

    if output_results:
        print_comparison_results("Model Output PSNR", output_results)

    return layer_results, output_results



# Copyright (C) 2025 Embedl AB

"""
Module containing a unified set of helper texts for CLI commands.
"""

# To avoid slowing down the CLI, avoid imports in this file.

CONFIG_HELPER = (
    "Path to an optional YAML configuration file for advanced settings."
)
DEVICE_HELPER = (
    "Target device name for deployment. "
    "Use command `list-devices` to view all available options."
)
OUTPUT_FILE_HELPER = "Path to the output file or directory where the resulting model will be saved."
SIZE_HELPER = "Input image size in format HEIGHT,WIDTH (e.g., 224,224)."



# Copyright (C) 2025 Embedl AB
"""
Test cases for the embedl-hub CLI init command.

"""

import itertools

import pytest
import yaml
from typer.testing import CliRunner

from embedl_hub.cli.init import init_cli


class DummyCtx:
    def __init__(self, id, name):
        self.id = id
        self.name = name


@pytest.fixture(autouse=True)
def mock_tracking_and_ctx(monkeypatch, tmp_path):
    """
    Pytest fixture to mock tracking functions and redirect the context file
    to a temporary path. This prevents tests from making real API calls and
    from interfering with the user's actual context file.
    """
    # Mock tracking functions
    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.set_project",
        lambda name: DummyCtx(f"dummy_project_id_{name}", name),
    )
    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.set_experiment",
        lambda name: DummyCtx(f"dummy_experiment_id_{name}", name),
    )
    monkeypatch.setattr("embedl_hub.cli.utils.assert_api_config", lambda: None)

    # Monkeypatch CTX_FILE to use a temporary file
    temp_ctx_file = tmp_path / ".embedl_hub_ctx"
    monkeypatch.setattr("embedl_hub.core.context.CTX_FILE", temp_ctx_file)


runner = CliRunner()


def test_init_new_project(tmp_path):
    """Test creating a new project with the -p flag."""
    result = runner.invoke(init_cli, ["init", "-p", "MyProject"])
    assert result.exit_code == 0
    assert "✓ Project:" in result.output
    assert "✓ Experiment:" in result.output
    ctx_file = tmp_path / ".embedl_hub_ctx"
    ctx = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    assert ctx["project_name"] == "MyProject"
    assert ctx["project_id"] == "dummy_project_id_MyProject"
    assert (
        ctx["experiment_name"].startswith("experiment_")
        or ctx["experiment_name"]
    )
    assert ctx["experiment_id"].startswith("dummy_experiment_id_")


def test_init_new_experiment(tmp_path):
    """Test creating a new experiment with the -e flag in an existing project."""
    runner.invoke(init_cli, ["init", "-p", "MyProject"])
    result = runner.invoke(init_cli, ["init", "-e", "MyExperiment"])
    assert result.exit_code == 0
    assert "✓ Experiment:" in result.output
    ctx_file = tmp_path / ".embedl_hub_ctx"
    ctx = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    assert ctx["experiment_name"] == "MyExperiment"
    assert ctx["experiment_id"] == "dummy_experiment_id_MyExperiment"


def test_init_no_project_error():
    """Test when creating an experiment without an initialized project leads to a new experiment."""
    result = runner.invoke(init_cli, ["init", "-e", "MyExperiment"])
    assert result.exit_code == 0
    assert "No active project, creating a new one" in result.output


def test_init_default_project_and_experiment(tmp_path):
    """Test creating a default project and experiment with no flags."""
    result = runner.invoke(init_cli, ["init"])
    assert result.exit_code == 0
    assert "✓ Project:" in result.output
    assert "✓ Experiment:" in result.output
    ctx_file = tmp_path / ".embedl_hub_ctx"
    ctx = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    assert ctx["project_id"].startswith("dummy_project_id_project_")
    assert ctx["experiment_id"].startswith("dummy_experiment_id_experiment_")
    assert ctx["project_name"].startswith("project_")
    assert ctx["experiment_name"].startswith("experiment_")


def test_switch_project_resets_experiment(tmp_path):
    """Switching project should reset experiment context."""
    runner.invoke(init_cli, ["init", "-p", "Proj1", "-e", "Exp1"])
    ctx_file = tmp_path / ".embedl_hub_ctx"
    ctx1 = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    runner.invoke(init_cli, ["init", "-p", "Proj2"])
    ctx2 = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    assert ctx2["project_name"] == "Proj2"
    assert ctx2["project_id"] == "dummy_project_id_Proj2"
    assert ctx2["project_id"] != ctx1["project_id"]
    assert ctx2["experiment_id"] != ctx1["experiment_id"]


def test_show_command_outputs_context():
    """Test the show command prints the current context."""
    runner.invoke(init_cli, ["init", "-p", "ShowProj", "-e", "ShowExp"])
    result = runner.invoke(init_cli, ["show"])
    assert result.exit_code == 0
    assert ".embedl_hub_ctx" in result.output
    assert "project_id" in result.output
    assert "experiment_id" in result.output
    assert "ShowProj" in result.output
    assert "ShowExp" in result.output


def test_ctx_file_created_and_updated(tmp_path):
    """Test that the context file is created and updated as expected."""
    ctx_file = tmp_path / ".embedl_hub_ctx"
    assert not ctx_file.exists()
    runner.invoke(init_cli, ["init", "-p", "FileTestProj"])
    assert ctx_file.exists()
    ctx = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    assert ctx["project_name"] == "FileTestProj"
    assert ctx["project_id"] == "dummy_project_id_FileTestProj"
    runner.invoke(init_cli, ["init", "-e", "FileTestExp"])
    ctx2 = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    assert ctx2["experiment_name"] == "FileTestExp"
    assert ctx2["experiment_id"] == "dummy_experiment_id_FileTestExp"


def test_init_always_creates_new_project_and_experiment(monkeypatch, tmp_path):
    """Test that running 'init' with no flags always creates a new project and experiment."""

    project_counter = itertools.count()
    experiment_counter = itertools.count()

    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.set_project",
        lambda name: DummyCtx(
            f"dummy_project_id_{next(project_counter)}", name
        ),
    )
    monkeypatch.setattr(
        "embedl_hub.core.tracking_utils.set_experiment",
        lambda name: DummyCtx(
            f"dummy_experiment_id_{next(experiment_counter)}", name
        ),
    )
    # First run: create initial context
    result1 = runner.invoke(init_cli, ["init"])
    assert result1.exit_code == 0
    ctx_file = tmp_path / ".embedl_hub_ctx"
    ctx1 = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    project_id_1 = ctx1["project_id"]
    experiment_id_1 = ctx1["experiment_id"]
    # Second run: should create a new project and experiment, not reuse the old ones
    result2 = runner.invoke(init_cli, ["init"])
    assert result2.exit_code == 0
    ctx2 = yaml.safe_load(ctx_file.read_text(encoding="utf-8"))
    project_id_2 = ctx2["project_id"]
    experiment_id_2 = ctx2["experiment_id"]
    assert project_id_2 != project_id_1, (
        "Project ID should change on each init with no flags"
    )
    assert experiment_id_2 != experiment_id_1, (
        "Experiment ID should change on each init with no flags"
    )


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2025 Embedl AB
"""
embedl-hub list-devices - List all available target devices for the commands
`compile`, `quantize` and `benchmark`.
"""

import typer

list_devices_cli = typer.Typer()


@list_devices_cli.command("list-devices")
def list_devices_command():
    """
    List all available target devices.

    A device name is used as input to the `--device` option
    in the commands `compile`, `quantize` and `benchmark`.
    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.core.hardware.qualcomm_ai_hub import print_device_table
    # pylint: enable=import-outside-toplevel

    print_device_table()



# Copyright (C) 2025 Embedl AB

"""
Utils for Qualcomm AI Hub integration.
"""

import qai_hub as hub
import qai_hub.public_api_pb2 as hub_pb
import qai_hub.public_rest_api as hub_api
from qai_hub.client import Client
from qai_hub.hub import _global_client

QAI_HUB_RUNTIME_NAMES = ["ONNX Runtime", "TensorFlow Lite", "QNN"]


def get_global_qai_hub_client() -> Client:
    """Get the global Qualcomm AI Hub API client."""
    return _global_client


def get_job_result(
    job_id: str, client_config: hub_api.ClientConfig
) -> hub_pb.JobResult:
    """Get the protobuf representation of a job result from Qualcomm AI Hub.

    Used to retrieve information about a job that is otherwise not exposed by
    the `qai_hub` package.
    """
    # pylint: disable-next=protected-access
    job_result: hub_pb.JobResult = hub.client._api_call(
        hub_api.get_job_results, client_config, job_id
    )
    return job_result


def parse_runtime_info(job_result: hub_pb.JobResult) -> str:
    """Extract the runtime name from job result protobuf.

    If no recognized runtime is found, an error will be raised.
    """

    runtime = None

    job_type = job_result.WhichOneof("result")

    job_runtime_names = []

    if job_type == "compile_job_result":
        job_runtime_names = [
            tool.name
            for tool in job_result.compile_job_result.compile_detail.tool_versions
        ]
    elif job_type == "profile_job_result":
        job_runtime_names = [
            runtime.name
            for runtime in job_result.profile_job_result.profile.runtime_config
        ]
    else:
        raise RuntimeError(f"Unrecognized job type: {job_type}.")

    for expected_runtime_name in QAI_HUB_RUNTIME_NAMES:
        if expected_runtime_name in job_runtime_names:
            runtime = expected_runtime_name
            break

    if runtime is None:
        raise RuntimeError(
            f"No recognized runtime in job result: {job_runtime_names}. "
            f"Expected one of: {QAI_HUB_RUNTIME_NAMES}."
        )

    return runtime



# Copyright (C) 2025 Embedl AB
"""
On-device profiling of models via Qualcomm AI Hub.
This module provides functionality to profile model latency and memory usage
on a target device, returning detailed execution statistics.
"""

from numbers import Number
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Tuple

import qai_hub as hub
import tabulate

from embedl_hub.core.hardware.qualcomm_ai_hub import create_device
from embedl_hub.core.hub_logging import console
from embedl_hub.core.onnx_utils import maybe_package_onnx_folder_to_file
from embedl_hub.core.qai_hub_utils import (
    get_global_qai_hub_client,
    get_job_result,
    parse_runtime_info,
)
from embedl_hub.core.tracking_utils import experiment_context
from embedl_hub.tracking import RunType, log_metric, log_param


class ProfileError(RuntimeError):
    """Raised when Qualcomm AI Hub profile job fails or times out."""


def _to_ms(val: Number):
    "Always return milliseconds as float, or None if not available."
    return float(val) / 1000.0 if val is not None else None


def _count_layers_by_unit(execution_detail: List[dict]) -> Dict[str, int]:
    """Count layers by compute unit (CPU, GPU, NPU) from execution detail."""
    counts = {"CPU": 0, "GPU": 0, "NPU": 0}
    for layer in execution_detail:
        unit = layer.get("compute_unit")
        if unit in counts:
            counts[unit] += 1
    return counts


def _to_megabytes(val: Number):
    "Convert bytes to megabytes as float, or None if not available."
    return float(val) / (1024 * 1024) if val is not None else None


def _log_metrics(summary: dict) -> None:
    """Log profiling metrics to the tracking system."""
    log_metric(
        "$latency",
        summary.get("mean_ms"),
    )
    log_metric(
        "$peak_memory_usage",
        summary.get("peak_memory_usage_mb"),
    )
    for unit, count in summary.get("layers_by_unit", {}).items():
        log_metric(
            f"$num_layers_{unit.lower()}",
            count,
        )


def _log_layers(details: list) -> None:
    """Log layer information to the tracking system."""

    for idx, layer in enumerate(details):
        log_param(f"$layer_name_{idx}", layer["name"])
        log_param(f"$layer_type_{idx}", layer["type"])
        log_param(f"$layer_compute_unit_{idx}", layer["compute_unit"])
        log_metric("$latency_per_layer", layer["execution_time"], idx)
        log_metric("$cycles_per_layer", layer["execution_cycles"], idx)


def _collect_top_n_layer_times(
    execution_detail, num: int
) -> List[Tuple[float, str, str]]:
    """
    Collect layer execution times for the top n slowest layers from the profile detail.

    Returns a list of tuples (time_ms, layer_name, layer_type) in descending order of time.
    """
    all_layer_times = [
        (_to_ms(layer["execution_time"]), layer["name"], layer["type"])
        for layer in execution_detail
    ]
    return sorted(all_layer_times, reverse=True)[:num]


def _make_layer_names_unique(execution_detail: List[dict]) -> None:
    """Modify execution_detail in-place to ensure all layer names are unique."""

    class LayerNamer:
        """Helper class to generate unique layer names."""

        def __init__(self, layer_details: dict):
            self.all_names_unique = True
            self.seen_names = set()
            for layer in layer_details:
                name = layer.get("name")
                if name in self.seen_names or name == "":
                    self.all_names_unique = False
                    break
                self.seen_names.add(name)

            self.counts: Dict[str, int] = {}

        def __call__(self, layer_name: str, layer_type: str) -> str:
            if self.all_names_unique and layer_name:
                return layer_name
            count = self.counts.get(layer_type, 0)
            new_layer_name = f"{layer_type.lower()}_{count}"
            self.counts[layer_type] = count + 1
            return new_layer_name

    layer_namer = LayerNamer(execution_detail)

    for layer in execution_detail:
        layer["name"] = layer_namer(layer["name"], layer["type"])


def profile_model(
    model: Path,
    device: str,
    project_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
) -> tuple[dict, dict]:
    """
    Profile model latency on a target device using Qualcomm AI Hub.
    Returns (summary_dict, full_profile_dict).
    Raises ProfileError on failure.
    """
    with experiment_context(
        project_name=project_name,
        experiment_name=experiment_name,
        run_type=RunType.BENCHMARK,
    ):
        hub_device = create_device(device)
        log_param("$device", device)

        with TemporaryDirectory() as tmpdir:
            model = maybe_package_onnx_folder_to_file(model, tmpdir)
            try:
                job = hub.submit_profile_job(
                    model=model,
                    device=hub_device,
                )
            except Exception as exc:
                raise ProfileError("Failed to submit profile job.") from exc

        log_param("$qai_hub_job_id", job.job_id)

        try:
            prof = job.download_profile()
        except Exception as exc:
            raise ProfileError("Failed to download profile.") from exc

        try:
            job_result = get_job_result(
                job.job_id, get_global_qai_hub_client().config
            )
            runtime = parse_runtime_info(job_result)
        except RuntimeError as exc:
            raise ProfileError(
                "Failed to extract runtime info from job."
            ) from exc
        log_param("$runtime", runtime)

        summary = prof.get("execution_summary", {})
        execution_detail = prof.get("execution_detail", [])
        _make_layer_names_unique(execution_detail)
        layer_counts = _count_layers_by_unit(execution_detail)
        layer_times = _collect_top_n_layer_times(execution_detail, 5)
        summary_dict = {
            "mean_ms": _to_ms(summary.get("estimated_inference_time")),
            "peak_memory_usage_mb": _to_megabytes(
                summary.get("estimated_inference_peak_memory")
            ),
            "layer_times": layer_times,
            "layers": prof.get("layers", []),
            "layers_by_unit": layer_counts,
        }
        _log_metrics(summary_dict)

        _log_layers(execution_detail)
        return summary_dict, prof


def _make_layer_times_table(layer_times: list) -> str:
    """Create a pretty-printed table of layer execution times as a string, with a title."""
    headers = ["Name", "Type", "Time (ms)"]
    rows = [
        [name, layer_type, f"{time:.2f}" if time is not None else "—"]
        for time, name, layer_type in layer_times
    ]
    title = f"Layer execution times (Top {len(layer_times)})"
    table = tabulate.tabulate(rows, headers=headers, tablefmt="github")
    return f"\n{title}\n{table}\n"


def print_profile_summary(summary: dict) -> None:
    """Print latency summary to the user in a consistent way."""
    if summary.get("mean_ms") is not None:
        console.print(f"[green]✓ Mean latency:[/] {summary['mean_ms']:.2f} ms")
    if summary.get("peak_memory_usage_mb") is not None:
        console.print(
            f"[green]✓ Peak memory usage:[/] {summary['peak_memory_usage_mb']:.2f} MB"
        )
    if summary.get("layers_by_unit"):
        units = summary["layers_by_unit"]
        console.print(
            f"[green]✓ Layers by compute unit:[/] NPU={units['NPU']}, GPU={units['GPU']}, CPU={units['CPU']}"
        )
    if summary.get("layer_times"):
        table = _make_layer_times_table(summary["layer_times"])
        console.print(table)



# Copyright (C) 2025 Embedl AB

"""Base class for configuration management."""

import importlib
import os
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

import yaml
from jinja2 import Environment, StrictUndefined
from mergedeep import Strategy, merge
from pydantic import BaseModel

# Global variable to store default configurations (yaml files)
default_configs: Dict[str, str] = {
    "quantize": "quantization_config.yaml.j2",
}

T = TypeVar("T", bound="ExperimentConfig")


def load_defaults(file_name: str) -> Dict[str, Any]:
    """Load default configuration from a YAML file."""
    template_content = importlib.resources.read_text(
        "embedl_hub.core.default_configs", file_name
    )

    # Define the Jinja environment
    env = Environment(
        undefined=StrictUndefined,  # blows up if you reference a var you didn't supply
        trim_blocks=True,
        lstrip_blocks=True,
    )

    # Make env-variables visible in the environment
    env.globals.update(os.environ)

    rendered = env.from_string(template_content).render()
    return yaml.safe_load(rendered)


@contextmanager
def temp_img_size_env(env_vars: Dict[str, Any]):
    """Temporarily overwrite environment variables and undo on exit."""
    # stash old values
    old = {k: os.environ.get(k) for k in env_vars}
    # set new ones
    os.environ.update(env_vars)
    try:
        yield
    finally:
        # restore (or delete if not set before)
        for k, v in old.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v


def load_default_config_with_size(
    config_class: Type['ExperimentConfig'],
    size: Optional[str],
    default_config_name: str,
):
    """Load the default config, populating the config with the correct sizes.

    Sizes are either provided by the user in the cli or defined in a jinja2 template.
    The population is done by temporarily defining the sizes as env-variables and
    read by the reader.
    """
    if size is None:
        env_vars = {}
    else:
        h_and_w = size.split(',')
        if len(h_and_w) != 2:
            raise ValueError(
                "Expected `size` to be a comma separated set of two values, e.g. 224,224."
            )
        env_vars = {"IMG_H": h_and_w[0], "IMG_W": h_and_w[1]}
    with temp_img_size_env(env_vars):
        cfg = config_class.model_construct(
            **load_defaults(default_configs[default_config_name])
        )
        return cfg


class ExperimentConfig(BaseModel):
    """Base class for experiment configuration."""

    def to_yaml(self, path: Path) -> str:
        """Convert the configuration to a YAML string."""
        yaml_txt = yaml.dump(self.model_dump)
        if path is not None:
            path.write_text(yaml_txt)
        return yaml_txt

    @classmethod
    def from_yaml(cls: Type[T], path: Union[str, Path]) -> T:
        """Load the configuration from a YAML file."""
        return cls.model_construct(
            **yaml.safe_load(Path(path).read_text("utf-8"))
        )

    def merge_yaml(
        self: T, other: Optional[Union[str, Path]], **override
    ) -> T:
        """Merge another YAML file into the current configuration."""
        return self.merge_dict(
            yaml.safe_load(Path(other).read_text("utf-8")) if other else None,
            **override,
        )

    def merge_dict(
        self: T, other: Optional[Mapping[str, Any]], **override
    ) -> T:
        """
        Merge *other* (usually parsed YAML) and **override (kwargs) into a copy.
        Precedence: kwargs > other > self.
        """
        base = self.model_dump()
        if other is not None:
            merged = merge(base, other, strategy=Strategy.REPLACE)
        else:
            merged = base
        merged = merge(
            merged, override, strategy=Strategy.REPLACE
        )  # kwargs win last
        return self.__class__.model_construct(**merged).cast_str_to_path()

    def cast_str_to_path(self: T) -> T:
        """
        Casts string attributes to Path objects if the type hint is Path.

        This is useful when loading from sources that don't distinguish
        between strings and paths, like environment variables.
        """
        for field_name, field_info in self.model_fields.items():
            value = getattr(self, field_name)
            if isinstance(value, str):
                # Check if the field's type annotation is Path or Optional[Path]
                # This is a simplified check. For more complex types like
                # Union[Path, int], a more robust check would be needed.
                is_path_type = field_info.annotation is Path
                if not is_path_type:
                    origin = get_origin(field_info.annotation)
                    if origin:  # Handles Union, Optional, etc.
                        args = get_args(field_info.annotation)
                        if Path in args:
                            is_path_type = True

                if is_path_type:
                    setattr(self, field_name, Path(value))
        return self

    def validate_config(self) -> None:
        """Validate the configuration."""
        self.model_validate(self.model_dump())



# Copyright (C) 2025 Embedl AB

"""Module for quantizing ONNX models using Qualcomm AI Hub."""

from __future__ import annotations

import zipfile
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Tuple, Union

import numpy as np
import qai_hub as hub

from embedl_hub.core.hub_logging import console
from embedl_hub.core.onnx_utils import (
    load_onnx_model,
    maybe_package_onnx_folder_to_file,
)
from embedl_hub.core.quantization.psnr import measure_psnr_between_models
from embedl_hub.core.quantization.quantization_config import QuantizationConfig
from embedl_hub.core.tracking_utils import experiment_context
from embedl_hub.tracking import RunType, log_metric, log_param


class QuantizationError(RuntimeError):
    """Raised when Qualcomm AI Hub quantization job fails or times out."""


@dataclass
class QuantizationResult:
    """Result of a quantization job."""

    model_path: Path
    job_id: str


def get_input_shapes(model_path: Path) -> Dict[str, Tuple[int, ...]]:
    """
    Get the input shape of the ONNX model.

    Args:
        model_path: Path to the ONNX model.

    Returns:
        Dictionary with input names and their shapes.
    """

    onnx_model = load_onnx_model(model_path)
    input_shapes = {}
    for model_input in onnx_model.graph.input:
        shape = []
        for dim in model_input.type.tensor_type.shape.dim:
            if dim.dim_value > 0:
                shape.append(dim.dim_value)
            else:
                shape.append(1)  # Use 1 for dynamic dimensions
        input_shapes[model_input.name] = tuple(shape)
    return input_shapes


def _generate_random_data(
    config: QuantizationConfig,
) -> Dict[str, List[np.ndarray]]:
    """
    Generate random calibration data for quantization.

    Args:
        config: QuantizationConfig.

    Returns:
        Dictionary with random calibration data.
    """
    config.num_samples = 1

    def _make_random_sample(shape: Tuple[int, ...]) -> np.ndarray:
        """Generate a random sample with the given shape."""
        return np.random.rand(*shape).astype(np.float32)

    inputs_and_shapes = get_input_shapes(config.model)

    return {
        input_name: [_make_random_sample(shape)]
        for input_name, shape in inputs_and_shapes.items()
    }


def make_calibration_dataset(
    config: QuantizationConfig,
) -> Dict[str, List[np.ndarray]]:
    """
    Create a Dataset for calibration data.

    Args:
        config: QuantizationConfig containing data_path.

    If single input, data path should contain numpy files (.npy).
    If multiple inputs, data path should contain subdirectories named after
    the input names, each containing numpy files (.npy).

    Returns:
        Dataset for calibration data.
    """
    data_path = config.data_path
    if not data_path or not data_path.is_dir():
        raise ValueError(f"Invalid data path: {data_path}")

    onnx_model = load_onnx_model(config.model)
    input_names = [i.name for i in onnx_model.graph.input]

    if len(input_names) == 1:
        # Single input model
        input_name = input_names[0]
        npy_files = sorted(list(data_path.glob("*.npy")))
        if not npy_files:
            raise FileNotFoundError(f"No .npy files found in {data_path}")
        return {input_name: [np.load(f) for f in npy_files]}

    # Multiple input model
    datasets = {}
    num_input_samples = set()
    for input_name in input_names:
        input_dir = data_path / input_name
        if not input_dir.is_dir():
            raise FileNotFoundError(
                f"Input directory not found for input '{input_name}': {input_dir}"
            )
        npy_files = sorted(list(input_dir.glob("*.npy")))
        if not npy_files:
            raise FileNotFoundError(f"No .npy files found in {input_dir}")
        datasets[input_name] = [np.load(f) for f in npy_files]
        num_input_samples.add(len(datasets[input_name]))

    if len(num_input_samples) != 1:
        raise ValueError("All inputs must have the same number of samples.")
    return datasets


def _load_calibration_data(
    config: QuantizationConfig,
) -> Dict[str, List[np.ndarray]]:
    """
    Load calibration data from the specified path.

    Args:
        config: QuantizationConfig containing data_path and transforms.

    Returns:
        Dictionary with calibration data.
    """

    dataset = make_calibration_dataset(config)
    for input_name, input_samples in dataset.items():
        dataset[input_name] = input_samples[: config.num_samples]
        config.num_samples = min(config.num_samples, len(dataset[input_name]))

    return dataset


def collect_calibration_data(
    config: QuantizationConfig,
) -> Dict[str, List[np.ndarray]]:
    """
    Collect calibration data for quantization.

    Args:
        config: QuantizationConfig containing data_path and transforms.

    Returns:
        Dictionary with calibration data.
    """
    if config.data_path is None:
        console.print(
            "[yellow]No data path specified, generating random calibration data.[/]"
        )
        return _generate_random_data(config)
    return _load_calibration_data(config)


def _save_quantized(quantized: "hub.Model", output_file: Path) -> Path:
    """
    Save the quantized artifact to a user-specified location or to the root
    directory with the name provided by Qualcomm AI Hub.

    Returns:
        Path to the saved model file.
    """
    output_file.parent.mkdir(parents=True, exist_ok=True)
    output_file = Path(quantized.download(str(output_file)))
    return output_file


def log_config_params(config: QuantizationConfig) -> None:
    """Log the configuration parameters for tracking."""

    def _format_img_size(value: Union[int, List[int]]) -> str:
        """Format image size for logging."""
        return str(value).strip("[]").replace(", ", "x").strip()

    log_param("num samples", str(config.num_samples))

    def _find_input_sizes(
        config: QuantizationConfig,
    ) -> List[Tuple[str, List[int]]]:
        """Find input sizes from the model."""
        onnx_model = load_onnx_model(config.model)
        input_sizes = []
        for model_input in onnx_model.graph.input:
            shape = []
            for dim in model_input.type.tensor_type.shape.dim:
                if dim.dim_value > 0:
                    shape.append(dim.dim_value)
                else:
                    shape.append(1)  # Use 1 for dynamic dimensions
            input_sizes.append((model_input.name, shape))
        return input_sizes

    for input_name, shape in _find_input_sizes(config):
        log_param(f"{input_name} input shape", _format_img_size(shape))


def _log_psnr_results(layer_psnr: List[Dict], output_psnr: List[Dict]) -> None:
    """Log PSNR results to the tracking system."""
    for idx, layer in enumerate(layer_psnr):
        log_param(f"$layer_name_{idx}", layer["layer"])
        log_metric("$psnr_per_layer", layer["psnr"], idx)
        layer_shape = " ".join(str(x) for x in layer["shape"]) or "1"
        log_param(f"$layer_shape_{idx}", layer_shape)
        if "notes" in layer:
            log_param(f"$layer_notes_{idx}", layer["notes"])

    for idx, output in enumerate(output_psnr):
        log_param(f"$output_name_{idx}", output["layer"])
        log_metric(f"$output_psnr_{idx}", output["psnr"])
        output_shape = " ".join(str(x) for x in output["shape"]) or "1"
        log_param(f"$output_shape_{idx}", output_shape)
        if "notes" in output:
            log_param(f"$output_notes_{idx}", output["notes"])


def _maybe_unzip_model(model_path: Path, tmp_folder: Path) -> Path:
    """Unzip the model if it is a zip file and return the ONNX file path."""
    if model_path.suffix == ".zip":
        with zipfile.ZipFile(model_path, "r") as zip_ref:
            extract_path = tmp_folder / model_path.stem
            zip_ref.extractall(extract_path)
            subfolders = [f for f in extract_path.iterdir() if f.is_dir()]
            if len(subfolders) != 1:
                raise RuntimeError(
                    f"Expected exactly one folder in the zip, found: {len(subfolders)}"
                )
            only_folder = subfolders[0]
            onnx_files = list(only_folder.rglob("*.onnx"))
            if len(onnx_files) != 1:
                raise RuntimeError(
                    f"Expected exactly one .onnx file, found: {len(onnx_files)}"
                )
            return onnx_files[0]
    return model_path


def quantize_model(
    config: QuantizationConfig,
    project_name: str,
    experiment_name: str,
) -> None:
    """
    Submit an ONNX model to Qualcomm AI Hub and retrieve the compiled artifact.

    Args:
        config: QuantizationConfig containing model_path, data_path, and transforms.

    Returns:
        QuantizationResult with local path to compiled model.

    """

    with experiment_context(project_name, experiment_name, RunType.QUANTIZE):
        if not config.model.exists():
            raise ValueError(f"Model not found: {config.model}")

        calibration_data = collect_calibration_data(config)
        log_config_params(config)

        with TemporaryDirectory() as tmpdir:
            config.model = maybe_package_onnx_folder_to_file(
                config.model, tmpdir
            )
            try:
                job = hub.submit_quantize_job(
                    model=config.model.as_posix(),
                    weights_dtype=hub.QuantizeDtype.INT8,
                    activations_dtype=hub.QuantizeDtype.INT8,
                    calibration_data=calibration_data,
                )
            except Exception as error:
                raise QuantizationError(
                    "Failed to submit quantization job."
                ) from error

            log_param("$qai_hub_job_id", job.job_id)

            try:
                quantized = job.get_target_model()
            except Exception as error:
                raise QuantizationError(
                    "Failed to download quantized model from Qualcomm AI Hub."
                ) from error
            if quantized is None:
                raise QuantizationError(
                    "Quantized model returned by Qualcomm AI Hub is None."
                )

            layer_psnr = []
            output_psnr = []

            local_path = _save_quantized(quantized, config.output_file)
            try:
                with TemporaryDirectory() as tmpdir:
                    tmpdir_path = Path(tmpdir)
                    layer_psnr, output_psnr = measure_psnr_between_models(
                        config.model,
                        _maybe_unzip_model(local_path, tmpdir_path),
                        calibration_data,
                    )
            except Exception as error:
                pass  # PSNR measurement is optional, ignore errors

        _log_psnr_results(layer_psnr, output_psnr)

        console.print(f"[green]✓ Quantized model saved to {local_path}[/]")



# Copyright (C) 2025 Embedl AB
"""
Test cases for the embedl-hub CLI init command.

"""

from unittest.mock import patch

import pytest
from typer.testing import CliRunner

from embedl_hub.cli.auth import auth_cli

runner = CliRunner()


@patch("embedl_hub.core.context.write_ctx")
@patch("embedl_hub.core.context.read_embedl_hub_context")
@patch("embedl_hub.core.hub_logging.console")
def test_auth_command_success(mock_console, mock_read_ctx, mock_write_ctx):
    """Test that auth_command successfully stores a new API key."""
    # Arrange
    mock_read_ctx.return_value = {}
    api_key = "test-api-key-123"

    # Act
    result = runner.invoke(auth_cli, ["--api-key", api_key])

    # Assert
    assert result.exit_code == 0
    mock_read_ctx.assert_called_once()
    mock_write_ctx.assert_called_once_with({"api_key": api_key})
    mock_console.print.assert_called_once_with("[green]✓ Stored API key[/]")


@patch("embedl_hub.core.context.write_ctx")
@patch("embedl_hub.core.context.read_embedl_hub_context")
@patch("embedl_hub.core.hub_logging.console")
def test_auth_command_updates_existing_key(
    mock_console, mock_read_ctx, mock_write_ctx
):
    """Test that auth_command successfully updates an existing API key."""
    # Arrange
    mock_read_ctx.return_value = {
        "api_key": "old-key",
        "other_data": "persists",
    }
    new_api_key = "new-api-key-456"

    # Act
    result = runner.invoke(auth_cli, ["--api-key", new_api_key])

    # Assert
    assert result.exit_code == 0
    mock_read_ctx.assert_called_once()
    mock_write_ctx.assert_called_once_with(
        {"api_key": new_api_key, "other_data": "persists"}
    )
    mock_console.print.assert_called_once_with("[green]✓ Stored API key[/]")


def test_auth_command_missing_api_key():
    """Test that the command fails if the --api-key option is missing."""
    # Act
    result = runner.invoke(auth_cli, [])

    # Assert
    assert result.exit_code != 0


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2025 Embedl AB


from embedl_hub.tracking.client import Client as _Client
from embedl_hub.tracking.rest_api import RunType

global_client = _Client()

set_project = global_client.set_project
set_experiment = global_client.set_experiment
start_run = global_client.start_run
log_param = global_client.log_param
log_metric = global_client.log_metric


__all__ = [
    "set_project",
    "set_experiment",
    "start_run",
    "log_param",
    "log_metric",
    "RunType",
    "global_client",
]



# Copyright (C) 2025 Embedl AB
"""
embedl-hub CLI

This is the main entry point for the Embedl-Hub command-line interface.
It provides commands for experiment tracking and on-device benchmarking
and verification.
"""

from importlib import metadata

import typer

from embedl_hub.cli.auth import auth_cli
from embedl_hub.cli.benchmark import benchmark_cli
from embedl_hub.cli.compile import compile_cli
from embedl_hub.cli.init import init_cli
from embedl_hub.cli.list_devices import list_devices_cli
from embedl_hub.cli.quantize import quantize_cli
from embedl_hub.core.hub_logging import console, setup_logging

app = typer.Typer(
    add_completion=True,
    rich_markup_mode="rich",
    help="[bold cyan]embedl-hub[/] end-to-end Edge-AI workflow CLI",
    no_args_is_help=True,
    # Disable local variables for security.
    # See https://typer.tiangolo.com/tutorial/exceptions/#disable-local-variables-for-security
    pretty_exceptions_show_locals=False,
)

# The order of these commands should be the expected order of calls,
# not alphabetical.
app.add_typer(auth_cli)
app.add_typer(init_cli)
app.add_typer(compile_cli)
app.add_typer(quantize_cli)
app.add_typer(benchmark_cli)
app.add_typer(list_devices_cli)


def _version_callback(value: bool):
    if value:
        console.print(f"embedl-hub {metadata.version('embedl-hub')}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-V",
        callback=_version_callback,
        is_eager=True,
        help="Print embedl-hub version and exit.",
    ),
    verbose: int = typer.Option(
        0,
        "--verbose",
        "-v",
        count=True,
        help="Increase verbosity (-v, -vv, -vvv).",
        show_default=False,
    ),
):
    """Main entry point for the embedl-hub CLI."""

    _ = version  # Mark as used to avoid unused argument warning
    setup_logging(verbose)



# Copyright (C) 2025 Embedl AB

"""
embedl-hub compile - send an model to Qualcomm AI Hub, retrieve a
device-specific binary (.tflite for tflite, .bin for qnn, or .onnx for onnxruntime).
"""

from pathlib import Path

import typer

# All other embedl_hub imports should be done inside the function.
from embedl_hub.cli.helper import (
    DEVICE_HELPER,
    OUTPUT_FILE_HELPER,
    SIZE_HELPER,
)

compile_cli = typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True,
)


@compile_cli.command("compile")
def compile_command(
    model: Path = typer.Option(
        ...,
        "-m",
        "--model",
        help="Path to the TorchScript model file, ONNX model file, or to a directory "
        "containing the ONNX model and any associated data files, to be compiled.",
        show_default=False,
    ),
    device: str = typer.Option(
        ...,
        "-d",
        "--device",
        help=DEVICE_HELPER,
        show_default=False,
    ),
    size: str = typer.Option(
        ...,
        "--size",
        "-s",
        help=SIZE_HELPER,
        show_default=False,
    ),
    runtime: str = typer.Option(
        ...,
        "-r",
        "--runtime",
        help="Runtime backend for compilation: tflite, qnn, or onnx.",
    ),
    quantize_io: bool = typer.Option(
        False,
        "--quantize-io",
        help="Quantize input and output tensors. "
        "Improves performance on platforms that support quantized I/O.",
        show_default=True,
    ),
    output_file: str = typer.Option(
        None,
        "-o",
        "--output-file",
        help=OUTPUT_FILE_HELPER,
        show_default=False,
    ),
):
    """
    Compile a model into a device ready binary using Qualcomm AI Hub.
    Qualcomm AI Hub may return a zip file containing multiple files.

    Required arguments:
        --model
        --size
        --device
        --runtime

    Examples
    --------

    Compile the ONNX model `fp32_model.onnx` with input size 1x3x224x224 for the Samsung Galaxy S24 using the tflite runtime:

        $ embedl-hub compile -m fp32_model.onnx  --size 1,3,224,224 -d "Samsung Galaxy S24" -r tflite

    Compile the TorchScript model `model.pt` with input size 1x3x224x224
    for the Samsung Galaxy S24, and save it to `./my_outputs/model.onnx`:

        $ embedl-hub compile -m model.pt  -r onnx --size 1,3,224,224 --device "Samsung Galaxy S24" -o ./my_outputs/model.onnx

    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import assert_api_config, prepare_input_size
    from embedl_hub.core.compile import (
        CompileError,
        CompileResult,
        compile_model,
    )
    from embedl_hub.core.context import require_embedl_hub_context
    from embedl_hub.core.hub_logging import console
    # pylint: enable=import-outside-toplevel

    assert_api_config()
    ctx = require_embedl_hub_context()

    if not model:
        raise ValueError("Please specify a model to compile using --model")
    if not size:
        raise ValueError(
            "Please specify input size using --size, e.g. 1,3,224,224"
        )

    if not output_file:
        output_file = model.with_suffix(f".{runtime}").as_posix()
        console.print(
            f"[yellow]No output file specified, using {output_file}[/]"
        )
    input_size = prepare_input_size(size)
    try:
        res: CompileResult = compile_model(
            project_name=ctx["project_name"],
            experiment_name=ctx["experiment_name"],
            model_path=model,
            device=device,
            runtime=runtime,
            quantize_io=quantize_io,
            output_file=output_file,
            input_size=input_size,
        )
        console.print(f"[green]✓ Compiled model for {res.device}[/]")

        # TODO: upload artifacts / metrics to web
    except (CompileError, ValueError) as error:
        console.print(f"[red]✗ {error}[/]")
        raise typer.Exit(1)



# Copyright (C) 2025 Embedl AB

"""Test the CLI for quantizing models using the Embedl Hub SDK."""

from pathlib import Path

import pytest
import yaml
from typer.testing import CliRunner

from embedl_hub.cli.quantize import quantize_cli


def test_quantize_cli_uses_default_config():
    """
    Test the CLI with default configuration.

    The test should fail due to lacking config settings.
    """

    runner = CliRunner()
    result = runner.invoke(quantize_cli)
    assert result.exit_code == 2
    assert isinstance(result.exception, SystemExit)


def test_quantize_cli_overrides_with_flags(monkeypatch):
    """Test the CLI with command-line flags to override default configuration."""

    captured = {}

    # pylint: disable-next=unused-argument
    def fake_quantize_model(config, project_name, experiment_name):
        """Fake quantize_model function to capture the config."""
        captured["cfg"] = config

    monkeypatch.setattr(
        "embedl_hub.core.quantization.quantize.quantize_model",
        fake_quantize_model,
    )
    monkeypatch.setattr('embedl_hub.cli.utils.assert_api_config', lambda: None)
    monkeypatch.setattr(
        "embedl_hub.core.context.read_embedl_hub_context",
        lambda: {
            "project_name": "test_project",
            "experiment_name": "test_experiment",
        },
    )

    runner = CliRunner()
    args = [
        "--model",
        "my-model",
        "--data",
        "/other/data",
    ]
    result = runner.invoke(quantize_cli, args)
    assert result.exit_code == 0

    cfg = captured["cfg"]
    assert cfg.model == Path("my-model")
    assert cfg.data_path == Path("/other/data")
    assert cfg.num_samples == 500  # default


def test_quantize_cli_with_custom_config_file(tmp_path, monkeypatch):
    """Test the CLI with a custom YAML configuration file."""

    # Create a custom YAML and pass via --config
    custom = {
        "data_path": "/mnt/x",
        "num_samples": 128,
    }
    custom_path = tmp_path / "custom.yaml"
    custom_path.write_text(yaml.dump(custom))

    captured = {}

    # pylint: disable-next=unused-argument
    def fake_quantize_model(config, project_name, experiment_name):
        """Fake quantize_model function to capture the config."""
        captured["cfg"] = config

    monkeypatch.setattr(
        "embedl_hub.core.quantization.quantize.quantize_model",
        fake_quantize_model,
    )
    monkeypatch.setattr('embedl_hub.cli.utils.assert_api_config', lambda: None)
    monkeypatch.setattr(
        "embedl_hub.core.context.read_embedl_hub_context",
        lambda: {
            "project_name": "test_project",
            "experiment_name": "test_experiment",
        },
    )

    runner = CliRunner()
    result = runner.invoke(
        quantize_cli, ["--model", "from-cli", "--config", str(custom_path)]
    )
    assert result.exit_code == 0

    cfg = captured["cfg"]
    assert cfg.model == Path("from-cli")
    assert cfg.data_path == Path("/mnt/x")
    assert cfg.num_samples == 128


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2025 Embedl AB

"""Module for ONNX related helper functions."""

from pathlib import Path
from typing import Union

import onnx

from embedl_hub.core.hub_logging import console


def load_onnx_model(model_path: Path) -> onnx.ModelProto:
    """Load an ONNX model from a file or directory.
    If a directory is provided, it expects exactly one ONNX file inside.
    Args:
        model_path: Path to the ONNX model file or directory.
    Returns:
        Loaded ONNX model.
    Raises:
        ValueError: If the model path is invalid or does not contain a valid ONNX file.
    """

    if model_path.is_file():
        if model_path.suffix.lower() != ".onnx":
            raise ValueError(f"Expected a .onnx file, got: {model_path.name}")
        return onnx.load(model_path)
    if not model_path.is_dir():
        raise ValueError(
            f"Model path must be a file or directory: {model_path}"
        )
    onnx_files = sorted([p for p in model_path.glob("*.onnx") if p.is_file()])
    if not len(onnx_files) == 1:
        raise ValueError(
            f"Expected exactly one ONNX file in directory, found: {len(onnx_files)}"
        )
    return onnx.load(onnx_files[0])


def maybe_package_onnx_folder_to_file(
    model_path: Path, dest_folder: Union[Path, str]
) -> Path:
    """If model_path is a directory, repackage the ONNX model into a single .onnx file.
    Returns the path to the .onnx file (either the original file or a new file inside dest_folder).
    """
    if model_path.is_dir():
        console.log("Directory model detected. Looking for ONNX file...")
        # Repackage directory onnx model into a single .onnx file
        onnx_model = load_onnx_model(model_path)
        new_onnx_path = (
            Path(dest_folder) / model_path.with_suffix(".onnx").name
        )
        onnx.save(onnx_model, new_onnx_path)
        console.log("ONNX model found and validated.")
        return new_onnx_path
    return model_path



# Copyright (C) 2025 Embedl AB

"""
CLI command to benchmark a model's performance on a device.

This command profiles the latency of a compiled model on a specified device
and saves the results in JSON format. Both a summary and a full profile
are generated and saved to the specified output directory. The full profile
includes detailed performance metrics useful for debugging.
"""

import json
from datetime import datetime
from pathlib import Path

import typer

# All other embedl_hub imports should be done inside the function.
from embedl_hub.cli.helper import DEVICE_HELPER

benchmark_cli = typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True,
)


@benchmark_cli.command("benchmark")
def benchmark_command(
    model: Path = typer.Option(
        ...,
        "-m",
        "--model",
        help="Path to a compiled model file (.tflite, .onnx, or .bin), or "
        "to a directory containing an ONNX model and its data files.",
        show_default=False,
    ),
    device: str = typer.Option(
        ..., "-d", "--device", help=DEVICE_HELPER, show_default=False
    ),
    output_dir: Path = typer.Option(
        None, "--output-dir", "-o", help="Output folder for benchmark JSONs"
    ),
):
    """Benchmark compiled model on device and measure it's performance.

    Examples:
    ---------
    Benchmark a .tflite model on Samsung Galaxy S25 and save results to
    the default benchmarks folder:

        $ embedl-hub benchmark -m my_model.tflite -d "Samsung Galaxy S25"

    Benchmark an .onnx model on Samsung Galaxy 8 Elite QRD and save
    results to a custom output directory:

        $ embedl-hub benchmark -m my_model.onnx -d "Samsung Galaxy 8 Elite QRD" -o results/

    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import assert_api_config
    from embedl_hub.core.benchmark import (
        ProfileError,
        print_profile_summary,
        profile_model,
    )
    from embedl_hub.core.context import require_embedl_hub_context
    from embedl_hub.core.hub_logging import console
    # pylint: enable=import-outside-toplevel

    assert_api_config()
    ctx = require_embedl_hub_context()

    console.log(f"profiling {model.name} on {device} using Qualcomm AI Hub")
    try:
        summary, full = profile_model(
            model,
            device,
            project_name=ctx["project_name"],
            experiment_name=ctx["experiment_name"],
        )
    except (ValueError, ProfileError) as e:
        console.print(f"[red]✗ profiling failed:[/] {e}")
        raise typer.Exit(1)
    print_profile_summary(summary)

    if output_dir is None:
        outdir = Path("benchmarks")
    else:
        outdir = Path(output_dir)
    outdir.mkdir(exist_ok=True)
    ts = datetime.now().strftime("%Y%m%d_%H%M%S")
    summary_path = outdir / f"{model.stem}_{ts}.json"
    full_path = outdir / f"{model.stem}_{ts}.full.json"

    with open(summary_path, "w", encoding="utf-8") as f:
        json.dump(summary, f, indent=2)
    console.print(f"[green]✓ Saved benchmark summary to:[/] {summary_path}")

    with open(full_path, "w", encoding="utf-8") as f:
        json.dump(full, f, indent=2)
    console.print(
        f"[green]✓ Saved full Qualcomm AI Hub benchmark to:[/] {full_path}"
    )



# Copyright (C) 2025 Embedl AB

"""
embedl-hub auth - authenticate with your API key
"""

from __future__ import annotations

import typer

auth_cli = typer.Typer(help="Initialise / show project & experiment context")


@auth_cli.command("auth")
def auth_command(
    api_key: str = typer.Option(
        ...,
        "--api-key",
        help="Set or update API key. Generate one at https://hub.embedl.com/profile.",
        show_default=False,
    ),
):
    """
    Store the API key for embedl-hub CLI.

    Examples
    --------
    Configure the API key:
        $ embedl-hub init --api-key <your-key>
    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.core.context import read_embedl_hub_context, write_ctx
    from embedl_hub.core.hub_logging import console
    # pylint: enable=import-outside-toplevel

    ctx = read_embedl_hub_context()
    ctx["api_key"] = api_key
    write_ctx(ctx)

    console.print("[green]✓ Stored API key[/]")



# Copyright (C) 2025 Embedl AB

"""Tests for the quantization module in the Embedl Hub SDK."""

from pathlib import Path

import numpy as np
import onnx
import pytest
from onnx import TensorProto, helper

from embedl_hub.core.quantization.quantization_config import QuantizationConfig
from embedl_hub.core.quantization.quantize import (
    _load_calibration_data,
    make_calibration_dataset,
)


def _create_onnx_model(model_path: Path, input_names: list[str]):
    """Helper function to create a dummy ONNX model."""
    inputs = [
        helper.make_tensor_value_info(name, TensorProto.FLOAT, [1, 3, 16, 16])
        for name in input_names
    ]
    output = helper.make_tensor_value_info(
        "output", TensorProto.FLOAT, [1, 10]
    )
    nodes = [
        helper.make_node("Add", [name, name], [f"add_{name}"])
        for name in input_names
    ]
    # Dummy node to consume all inputs
    final_node = helper.make_node(
        "Sum", [f"add_{name}" for name in input_names], ["output"]
    )

    graph_def = helper.make_graph(
        nodes + [final_node], "test-model", inputs, [output]
    )
    model_def = helper.make_model(graph_def, producer_name="test")
    onnx.save(model_def, model_path)


@pytest.fixture
def temp_path(tmp_path: Path) -> Path:
    """Fixture to create a temporary directory for tests."""
    return tmp_path


def test_single_input_model(temp_path: Path):
    """Test make_calibration_dataset with a single input model."""
    model_path = temp_path / "single_input.onnx"
    _create_onnx_model(model_path, ["input1"])

    data_path = temp_path / "cal_data"
    data_path.mkdir()
    for i in range(3):
        np.save(data_path / f"sample_{i}.npy", np.random.rand(1, 3, 16, 16))

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=3,
        output_file=temp_path / "out.onnx",
    )
    dataset = make_calibration_dataset(config)

    assert isinstance(dataset, dict)
    assert "input1" in dataset
    assert len(dataset["input1"]) == 3
    sample = dataset["input1"][0]
    assert sample.shape == (1, 3, 16, 16)


def test_multiple_input_model(temp_path: Path):
    """Test make_calibration_dataset with a multiple input model."""
    model_path = temp_path / "multi_input.onnx"
    input_names = ["input1", "input2"]
    _create_onnx_model(model_path, input_names)

    data_path = temp_path / "cal_data"
    data_path.mkdir()
    for name in input_names:
        input_dir = data_path / name
        input_dir.mkdir()
        for i in range(2):
            np.save(
                input_dir / f"sample_{i}.npy", np.random.rand(1, 3, 16, 16)
            )

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=2,
        output_file=temp_path / "out.onnx",
    )
    dataset = make_calibration_dataset(config)

    assert isinstance(dataset, dict)
    assert "input1" in dataset
    assert "input2" in dataset
    assert len(dataset["input1"]) == 2
    assert len(dataset["input2"]) == 2
    input1_sample = dataset["input1"][0]
    assert input1_sample.shape == (1, 3, 16, 16)
    input2_sample = dataset["input2"][0]
    assert input2_sample.shape == (1, 3, 16, 16)


def test_invalid_data_path(temp_path: Path):
    """Test error when data_path is not a directory."""
    model_path = temp_path / "model.onnx"
    _create_onnx_model(model_path, ["input1"])
    config = QuantizationConfig(
        model=model_path,
        data_path=temp_path / "non_existent",
        num_samples=2,
        output_file=temp_path / "out.onnx",
    )
    with pytest.raises(ValueError):
        make_calibration_dataset(config)


def test_missing_input_dir_for_multi_input(temp_path: Path):
    """Test error when an input directory is missing for a multi-input model."""
    model_path = temp_path / "multi_input.onnx"
    _create_onnx_model(model_path, ["input1", "input2"])

    data_path = temp_path / "cal_data"
    data_path.mkdir()
    (data_path / "input1").mkdir()  # Only create dir for input1

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=1,
        output_file=temp_path / "out.onnx",
    )
    with pytest.raises(FileNotFoundError):
        make_calibration_dataset(config)


def test_no_npy_files_found(temp_path: Path):
    """Test error when no .npy files are found in the data directory."""
    model_path = temp_path / "model.onnx"
    _create_onnx_model(model_path, ["input1"])

    data_path = temp_path / "cal_data"
    data_path.mkdir()

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=1,
        output_file=temp_path / "out.onnx",
    )
    with pytest.raises(FileNotFoundError):
        make_calibration_dataset(config)


def test_load_calibration_data_single_input(temp_path: Path):
    """Test _load_calibration_data with a single input model."""
    model_path = temp_path / "single_input.onnx"
    _create_onnx_model(model_path, ["input1"])

    data_path = temp_path / "cal_data"
    data_path.mkdir()
    for i in range(5):
        np.save(data_path / f"sample_{i}.npy", np.random.rand(1, 3, 16, 16))

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=3,
        output_file=temp_path / "out.onnx",
    )
    data = _load_calibration_data(config)

    assert "input1" in data
    assert len(data["input1"]) == 3
    assert config.num_samples == 3


def test_load_calibration_data_multi_input(temp_path: Path):
    """Test _load_calibration_data with a multi-input model."""
    model_path = temp_path / "multi_input.onnx"
    input_names = ["input1", "input2"]
    _create_onnx_model(model_path, input_names)

    data_path = temp_path / "cal_data"
    data_path.mkdir()
    for name in input_names:
        input_dir = data_path / name
        input_dir.mkdir()
        for i in range(5):
            np.save(
                input_dir / f"sample_{i}.npy", np.random.rand(1, 3, 16, 16)
            )

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=2,
        output_file=temp_path / "out.onnx",
    )
    data = _load_calibration_data(config)

    assert "input1" in data
    assert "input2" in data
    assert len(data["input1"]) == 2
    assert len(data["input2"]) == 2
    assert config.num_samples == 2


def test_load_calibration_data_num_samples_greater_than_available(
    temp_path: Path,
):
    """Test _load_calibration_data when num_samples is more than available."""
    model_path = temp_path / "single_input.onnx"
    _create_onnx_model(model_path, ["input1"])

    data_path = temp_path / "cal_data"
    data_path.mkdir()
    for i in range(3):
        np.save(data_path / f"sample_{i}.npy", np.random.rand(1, 3, 16, 16))

    config = QuantizationConfig(
        model=model_path,
        data_path=data_path,
        num_samples=10,
        output_file=temp_path / "out.onnx",
    )
    data = _load_calibration_data(config)

    assert "input1" in data
    assert len(data["input1"]) == 3
    assert config.num_samples == 3


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2025 Embedl AB

import os
from contextlib import contextmanager
from datetime import datetime, timezone

from embedl_hub.core.context import read_embedl_hub_context
from embedl_hub.tracking.rest_api import (
    ApiConfig,
    CompletedRunStatus,
    Experiment,
    Metric,
    Parameter,
    Project,
    Run,
    RunStatus,
    RunType,
    create_experiment,
    create_project,
    create_run,
    get_experiment_by_name,
    get_project_by_name,
    log_metric,
    log_param,
    update_run,
)

API_KEY_ENV_VAR_NAME = "EMBEDL_HUB_API_KEY"
BASE_URL_ENV_VAR_NAME = "EMBEDL_HUB_API_BASE_URL"

DEFAULT_API_BASE_URL = "https://hub.embedl.com/"


class Client:
    """Tracks projects, experiments and runs for the Embedl Hub web app."""

    _api_config: ApiConfig | None
    _project: Project | None
    _experiment: Experiment | None
    _active_run: Run | None

    def __init__(self, api_config: ApiConfig | None = None) -> None:
        self._api_config = api_config
        self._project = None
        self._experiment = None
        self._active_run = None

    def set_project(self, name: str) -> Project:
        """Set or create the current project by name."""

        project = get_project_by_name(self.api_config, name)

        if not project:
            project = create_project(self.api_config, name)

        self._project = project

        return project

    def set_experiment(self, name: str) -> Experiment:
        """Set or create the current experiment by name."""

        project_id = self.project.id
        experiment = get_experiment_by_name(self.api_config, name, project_id)

        if not experiment:
            experiment = create_experiment(self.api_config, name, project_id)

        self._experiment = experiment

        return experiment

    def create_run(self, type: RunType, name: str | None = None) -> Run:
        """Create a new run for the current project and experiment."""

        project = self.project
        experiment = self.experiment

        run = create_run(
            self.api_config,
            type=type,
            name=name,
            started_at=datetime.now(timezone.utc),
            project_id=project.id,
            experiment_id=experiment.id,
        )

        return run

    def update_active_run(
        self, status: CompletedRunStatus, ended_at: datetime
    ) -> None:
        """Update the status and end time of the active run."""

        project = self.project
        experiment = self.experiment
        run = self.active_run

        update_run(
            self.api_config,
            status=status,
            ended_at=ended_at,
            project_id=project.id,
            experiment_id=experiment.id,
            run_id=run.id,
        )

    @contextmanager
    def start_run(self, type: RunType, name: str | None = None):
        """Context manager to start and finish a run."""

        run = self.create_run(type, name)
        self._active_run = run

        status: CompletedRunStatus = RunStatus.FINISHED

        try:
            yield run
        except KeyboardInterrupt:
            status = RunStatus.KILLED
            raise
        except Exception:
            status = RunStatus.FAILED
            raise
        finally:
            self.update_active_run(
                status=status, ended_at=datetime.now(timezone.utc)
            )
            self._active_run = None

    def log_param(self, name: str, value: str) -> Parameter:
        """Log a parameter for the current run."""

        project = self.project
        experiment = self.experiment
        active_run = self.active_run

        param = log_param(
            self.api_config,
            name=name,
            value=value,
            project_id=project.id,
            experiment_id=experiment.id,
            run_id=active_run.id,
        )

        return param

    def log_metric(
        self, name: str, value: float, step: int | None = None
    ) -> Metric:
        """Log a metric for the current run."""

        project = self.project
        experiment = self.experiment
        active_run = self.active_run

        metric = log_metric(
            self.api_config,
            name=name,
            value=value,
            step=step,
            project_id=project.id,
            experiment_id=experiment.id,
            run_id=active_run.id,
        )

        return metric

    @property
    def api_config(self) -> ApiConfig:
        """Get or create the API config from environment variables."""

        if self._api_config is None:
            # environment variable takes precedence over context
            def get_api_key() -> str | None:
                """Get API key from environment or context."""
                if key := os.getenv(API_KEY_ENV_VAR_NAME):
                    return key
                return read_embedl_hub_context().get("api_key")

            if api_key := get_api_key():
                api_base_url = os.getenv(
                    BASE_URL_ENV_VAR_NAME, DEFAULT_API_BASE_URL
                )
                self._api_config = ApiConfig(
                    api_key=api_key, base_url=api_base_url
                )
            else:
                raise RuntimeError(
                    "No API key found. "
                    f"{API_KEY_ENV_VAR_NAME} must be set or stored in context by running `embedl-hub auth --api-key`."
                )

        return self._api_config

    @property
    def project(self) -> Project:
        if self._project is None:
            raise RuntimeError("Project is not set. Use set_project() first.")

        return self._project

    @property
    def experiment(self) -> Experiment:
        if self._experiment is None:
            raise RuntimeError(
                "Experiment is not set. Use set_experiment() first."
            )

        return self._experiment

    @property
    def active_run(self) -> Run:
        if self._active_run is None:
            raise RuntimeError(
                "There is no active run. Use start_run() as a context manager first."
            )

        return self._active_run



# Copyright (C) 2025 Embedl AB

"""Configuration for quantizing models."""

from pathlib import Path
from typing import Optional

from embedl_hub.core.config import ExperimentConfig


class QuantizationConfig(ExperimentConfig):
    """Class for quantization configuration."""

    # User specific parameters
    model: Path
    data_path: Optional[Path]
    output_file: Path
    num_samples: int
